-------------------------------------------------------------------------------
--	FILE:	Isolation
--	AUTHOR:	Snake
--	WEB:	SnakeByteStudios.com
-------------------------------------------------------------------------------

--[[

ABOUT

In this map script, each civilization and city-state is isolated in a small, fertile area only big enough for one major city. Between them is nothing but wasteland. This script is great for one-city challenges, exploration, or players that simply don't want to be right up against other civs. The script contain many options such as changing the terrain and geography of the wasteland or the size of starting areas.

I can guarantee that this map script is unlike any you've played before. Most maps generate terrain before finding player starts, but this one finds the starts and then generates terrain around them. Because of this, the entire generation process was written from scratch. And this makes for a very unique and interesting world.


OPTIONS

Dominant Terrain - Sets the terrain that the wasteland will be.

Wasteland Ridges - Toggles the mountain ranges that spread across the wasteland. Without them, the wasteland is even more desolate.

Player Size - Sets the tile radius of the civilization starts in the game. E.g. when this is set to 5, the starting areas will be 11 tiles across or 5 * 2 + 1.

City-State Size - Same as the previous option, but for city-states.

Resources - Similar to the generic resources option that most standard maps have, however completely customized. Sparse will yield about five resources, normal about nine, and abundant about thirteen.

Human Start Epicness - This option allows players to stack the odds in their favor by granting bonuses to only the human player. This option is disabled in multiplayer, however. At each level, more bonuses are added as such:
	Low - Gives player a couple more random resources
	Moderate - Gives player a couple more strategic resources and an extra ancient ruins + previous bonuses
	High - Gives player the best starting point and a couple more random resources + previous bonuses
	Epic - Gives player a natural wonder and all strategic resources + previous bonuses
	
Oddity Level - Adds interesting features to the wasteland. Higher levels increase both the oddness and the amount of oddities.

]]

include("MapGenerator")

------------------------------------------------------------------------------
function GetMapScriptInfo()
	local terrain = {
		Name = "TXT_KEY_MAP_OPTION_DOMINANT_TERRAIN", -- Dominant Terrain
		Values = {
			{"TXT_KEY_TERRAIN_SNOW"}, --Snow
			{"TXT_KEY_TERRAIN_DESERT"} --Desert
		},
		DefaultValue = 1,
		SortPriority = 1,
	};
	local ridges = {
		Name = "Wasteland Ridges",
		Values = {
			{"On"},
			{"Off"}
		},
		DefaultValue = 1,
		SortPriority = 2,
	};
	local player_size = {
		Name = "Player Size",
		Values = {
			{3},
			{4}, 
			{5}, 
		},
		DefaultValue = 3,
		SortPriority = 3,
	};
	local cs_size = {
		Name = "City-State Size",
		Values = {
			{2},
			{3}, 
		},
		DefaultValue = 1,
		SortPriority = 4,
	};
	local resources = {
		Name = "TXT_KEY_MAP_OPTION_RESOURCES", -- Resources
		Values = {
			{"TXT_KEY_MAP_OPTION_SPARSE"}, --Sparse
			{"TXT_KEY_MAP_OPTION_NORMAL"}, --Normal
			{"TXT_KEY_MAP_OPTION_ABUNDANT"}, --Abundant
		},
		DefaultValue = 2,
		SortPriority = 5,
	};
	local epicness = {
		Name = "Human Start Epicness",
		Values = {
			{"None"},
			{"TXT_KEY_MAP_OPTION_LOW"}, --Low
			{"TXT_KEY_MAP_OPTION_MEDIUM"}, --Medium
			{"TXT_KEY_MAP_OPTION_HIGH"}, --High
			{"TXT_KEY_PGSCREEN_SPEED_EPIC"}, --Epic
		},
		DefaultValue = 1,
		SortPriority = 6,
	};
	local oddity = {
		Name = "Oddity Level",
		Values = {
			{"None"},
			{"Curious"},
			{"Peculiar"},
			{"Ruddy Mysterious"},
			{"Weird"},
			{"WTF"}
		},
		DefaultValue = 2,
		SortPriority = 7,
	};
	return {
		Name = "Isolation",
		Description = "Each civilization and city-state is isolated in a small, fertile area only big enough for one major city. Between them is nothing but wasteland.",
		IsAdvancedMap = 0,
		IconIndex = 1,
		SortIndex = 1,
		CustomOptions = {
			terrain, --snow or desert
			ridges, --wasteland ridges
			player_size, --player biosphere size
			cs_size, --city state biosphere size
			resources, --resource abundance
			epicness, --player start epicness
			oddity --oddity level
		}
	};
end

local iW, iH, iNumCivs, iNumCityStates, player_ID_list, bTeamGame, teams_with_major_civs, number_civs_per_team
local isGAndK = false

--this is the map script entry point
--we're overriding the default version cause this shit is all custom
function GenerateMap()
	print("Generating Map")
	
	--do this first because almost everything depends on knowing where all the civs are going to start
	GenerateStarts()
	
	GeneratePlotTypes()
	
	GenerateTerrain()
	
	-- i don't really know what this does, but existing scripts have it
	Map.RecalculateAreas();
	
	--FixTerrainHoles()
	
	AddLakes()
	
	--the rivers depend on the lakes, so generate them afterwards
	AddRivers()
	
	AddFeatures()

	-- i don't really know what this does, but existing scripts have it
	Map.RecalculateAreas()

	--place starting points before resources so we know where the epic start is
	PlaceStartPlots()
	
	PlaceResources()
	
	PlaceNaturalWonders()

	AddGoodies()
	
	PlaceOddities()

	SetContinentArt()
end

--get the number of tiles that's in a hexagonal area based on its radius
function GetNumberOfTiles(radius)
	local tiles = 1
	for i = 1, radius do
		tiles = tiles + (i * 6)
	end
	return tiles
end

--a better random function than the weirdness Firaxis made
function GetRandom(lower, upper)
	if lower == nil then lower = 0 end
	if upper == nil then upper = 1 end
	return Map.Rand((upper + 1) - lower, "") + lower
end

--why the hell does Lua not have a round function
function Round(number)
	if number >= 0 then 
		return math.floor(number + .5)
	else 
		return math.ceil(number - .5)
	end
end

--get the plots that are between min and max distance from the plot at x and y
--returns {[i] = {x, y, depth}}
function GetPlotsAtDistance(x, y, min_distance, max_distance)
	local plots = {}
	local nextPlot = Map.GetPlot(x, y)
	
	--figure out if we're going to hit map boundaries by moving to min or max distance
	if y + max_distance > iH - 1 and y - max_distance < 0 then
		--this isn't going to work if we can't expand either N or S
		return nil
	end
	
	--if the min distance is 0, include the center plot
	if min_distance == 0 then
		table.insert(plots, {x = nextPlot:GetX(), y = nextPlot:GetY(), depth = 0})
		if max_distance == min_distance then return plots end
		min_distance = 1
	end
	
	--first move away from the center plot until you get to min distance
	for i = 1, min_distance do
		nextPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST)
		if nextPlot == nil then
			--just in case we hit map boundaries
			--although the math says we shouldn't
			return nil
		end
	end
	
	--if we were successful at moving to min distance
	--start looping around, collecting plots
	local direction
	for i = min_distance, max_distance do
		--then spiral around this depth
		direction = DirectionTypes.DIRECTION_SOUTHEAST
		for j = 1, 6 do
			--as you get farther out from center, you have to travel more in each direction
			for k = 1, i do
				nextPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), direction)
				if nextPlot == nil then return nil end
				table.insert(plots, {x = nextPlot:GetX(), y = nextPlot:GetY(), depth = i})
			end
			
			direction = (direction + 1) % 6
		end
		
		--if we need to collect more than one distance then jump to the next
		if i < max_distance then
			nextPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST)
			if nextPlot == nil then return nil end
		end
	end
	
	return plots
end

starting_points = {}
player_size_values = {3, 4, 5}
player_size = player_size_values[3]
cs_size_values = {2, 3}
cs_size = player_size_values[2]

epic_start = 1

is_worldbuilder = false

function GenerateStarts()
	iW, iH = Map.GetGridSize()
	iNumCivs, iNumCityStates, player_ID_list, bTeamGame, teams_with_major_civs, number_civs_per_team = GetPlayerAndTeamInfo()
	
	--if the game has crabs, then it's G&K. and a slut.
	if GameInfo.Resources.RESOURCE_CRAB ~= nil then isGAndK = true end
	
	--get the epic start option value for easy access and disable it if we're in multiplayer
	if Map.GetCustomOption(6) > 1 then
		if not (Game.IsGameMultiPlayer() or Game.IsHotSeat()) then
			epic_start = Map.GetCustomOption(6)
		else
			print ("Epic Start Disabled in Multiplayer")
		end
	end
	
	--for testing. civs will be 0 in worldbuilder.
	if iNumCivs == 0 then
		iNumCivs = 6
		iNumCityStates = 0
		is_worldbuilder = true
	end
	
	-- PLAYERS
	
	player_size = player_size_values[Map.GetCustomOption(3)]
	
	
	--must be 5 tiles away from the edge or errors will arise from terrain gen
	local edge_dist = player_size
	local min_player_dist = player_size + 2
	--basically the following gets the average radius that each start can have based on map size and number of players
	local player_dist = math.max(math.ceil(math.sqrt((iW * iH) / iNumCivs)/1.6), min_player_dist)
	local max_attempts = (iW * iH)
	print("Player Start Spacing: " .. player_dist .. " | Minimum: " .. min_player_dist)
	
	for i = 1, iNumCivs do
		--pick a random starting point
		local randX, randY, good_start
		for j = 1, max_attempts do
			
			randX = math.floor(Map.Rand(iW, ""))
			randY = math.ceil(edge_dist + Map.Rand(iH - (edge_dist * 2), ""))
			
			--print ("X: " .. randX)
			--print ("Y: " .. randY)
		
			--make sure it's not too close to other start points
			good_start = true
			for k=1, table.getn(starting_points) do
				if (Map.PlotDistance(starting_points[k].x, starting_points[k].y, randX, randY) < (player_dist + 1)) then
					good_start = false
					break
				end
			end
			
			if good_start == true then 
				print("Found Player " .. i .. " start: " .. randX .. ", " .. randY .. " | Attempts: " .. j .. " | Spacing: " .. player_dist)
				break
			elseif j % 50 == 0 then
				--every x attempts, reduce the distance threshold some
				if player_dist > min_player_dist then
					player_dist = math.max(math.floor(player_dist - (player_dist * .15)), min_player_dist)
				end
			end
			if j == max_attempts then
				max_attempts = 0
				break
			end
		end
		
		--add to list of other start points
		if good_start == true then
			--print ("X: " .. randX)
			--print ("Y: " .. randY)
			--print ("XY: " .. plotXY)
			table.insert(starting_points, {x = randX, y = randY, cityState = false, plots = {}})
		elseif max_attempts == 0 then
			print ("Failed to place anymore civs!")
			break
		end
	end
	
	--revise the number of civs in case any failed to place
	iNumCivs = table.getn(starting_points)
	
	-- CITY-STATES
	
	cs_size = cs_size_values[Map.GetCustomOption(4)]
	
	if iNumCityStates > 0 then
		--must be 3 tiles away from the edge or errors will arise from terrain gen
		edge_dist = cs_size
		max_attempts = (iW * iH)
		local min_cs_dist = cs_size + 2
		--the following gets the average radius that each city state start can have based on map size, number of players, and number of city states
		local cs_dist = math.max(math.ceil(math.sqrt(((iW * iH) - (iNumCivs * GetNumberOfTiles(player_size))) / iNumCityStates) / 2), min_cs_dist)
		print("CS Start Spacing: " .. cs_dist .. " | Minimum: " .. min_cs_dist)
		
		for i = 1, iNumCityStates do
			--pick a random starting point
			local randX, randY, good_start
			for j = 1, max_attempts do
				
				randX = math.floor(Map.Rand(iW, ""))
				randY = math.ceil(edge_dist + Map.Rand(iH - (edge_dist * 2), ""))
				
				--print ("X: " .. randX)
				--print ("Y: " .. randY)
			
				--make sure it's not too close to other start points
				good_start = true
				for k=1, table.getn(starting_points) do
					if (Map.PlotDistance(starting_points[k].x, starting_points[k].y, randX, randY) < (cs_dist + 1)) then
						good_start = false
						break
					end
				end
				
				if good_start == true then 
					print("Found CS " .. i .. " start: " .. randX .. ", " .. randY .. " | Attempts: " .. j .. " | Spacing: " .. cs_dist)
					break
				elseif j % 50 == 0 then
					--every x attempts, reduce the distance threshold some
					if cs_dist > min_cs_dist then
						cs_dist = math.max(math.floor(cs_dist - (cs_dist * .15)), min_cs_dist)
					end
				end
				if j == max_attempts then
					max_attempts = 0
					break
				end
			end
			
			--add to list of other start points
			if good_start == true then
				--print ("X: " .. randX)
				--print ("Y: " .. randY)
				--print ("XY: " .. plotXY)
				table.insert(starting_points, {x = randX, y = randY, cityState = true, plots ={}})
			elseif max_attempts == 0 then
				print ("Failed to place anymore city-states!")
				break
			end
		end
	end
	
	--revise the number of city states in case any failed to place
	iNumCityStates = table.getn(starting_points) - iNumCivs
	
	--go through all the plots and find which start they belong to
	--we'll need this for later
	for y = 0, iH - 1 do
		for x = 0, iW - 1 do
			for i = 1, table.getn(starting_points) do
				local start_size;
				if starting_points[i].cityState == false then
					start_size = player_size
				else
					start_size = cs_size
				end
				
				if (Map.PlotDistance(starting_points[i].x, starting_points[i].y, x, y) < (start_size + 1)) then
					local localX = math.abs((x - starting_points[i].x + start_size) % (iW - 1)) --have to account for wrapping
					local localY = y - starting_points[i].y + start_size
					table.insert(starting_points[i].plots, {
						x = x, 
						y = y, 
						localX = localX, 
						localY = localY,
						depth = Map.PlotDistance(starting_points[i].x, starting_points[i].y, x, y)
					})
					break
				end
			end
		end
	end
	
	--[[
	--debug
	for i = 1, table.getn(starting_points[1].plots) do
		print (starting_points[1].plots[i].x .. ", " .. starting_points[1].plots[i].y)
	end
	]]
end

function GeneratePlotTypes()
	--start with all land
	local plotTypes = {}
	for i = 0, Map.GetNumPlots() do
		plotTypes[i] = PlotTypes.PLOT_LAND
	end
	SetPlotTypes(plotTypes)
	
	local ridgesEnabled = Map.GetCustomOption(2)
	if ridgesEnabled == 1 then
		--generate some fractal mountain/hill ridges
		local numPlates = math.floor((iW * iH) / 200)
		print("Mountain Plates: " .. numPlates)
		local mountains = Fractal.Create(iW, iH, 4, {FRAC_WRAP_X = true}, -1, -1)
		mountains:BuildRidges(numPlates, {FRAC_WRAP_X = true}, 3, 1)
		local mountainThreshold = mountains:GetHeight(94)
		local hillThreshold = mountains:GetHeight(87)
		
		for y = 0, iH - 1 do
			for x = 0, iW - 1 do
				local height = mountains:GetHeight(x, y)
				local plot = Map.GetPlot(x, y)
				--print (x .. ", " .. y .. ": " .. height)
				if height >= mountainThreshold then
					plot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false);
				elseif height >= hillThreshold then
					plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
				end
			end
		end
	end
	
	--generate hills for each player start
	for i = 1, table.getn(starting_points) do
		local start_size;
		if starting_points[i].cityState == false then
			start_size = player_size
		else
			start_size = cs_size
		end
	
		local playerMountains = Fractal.Create((start_size * 2) + 1, (start_size * 2) + 1, 3, {}, -1, -1)
		local mountainThreshold = playerMountains:GetHeight(98)
		local hillThreshold = playerMountains:GetHeight(72)
		
		for j = 1, table.getn(starting_points[i].plots) do
			local x = starting_points[i].plots[j].x
			local y = starting_points[i].plots[j].y
			local localX = starting_points[i].plots[j].localX
			local localY = starting_points[i].plots[j].localY

			local height = playerMountains:GetHeight(localX, localY)
			local plot = Map.GetPlot(x, y)
			--print (x .. ", " .. y .. ": " .. height)
			if height >= mountainThreshold then
				plot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false);
			elseif height >= hillThreshold then
				plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
			else
				plot:SetPlotType(PlotTypes.PLOT_LAND, false, false);
			end
		end
	end
	
	--[[
	for i = 0, Map.GetNumPlots()-1 do
		print (Map.GetPlotByIndex(i):GetPlotType())
	end	
	]]
end

function GenerateTerrain()
	--set our base terrain
	local terrainTypes = {}
	local defaultTerrain
	if Map.GetCustomOption(1) == 1 then
		defaultTerrain = TerrainTypes.TERRAIN_SNOW
	else
		defaultTerrain = TerrainTypes.TERRAIN_DESERT
	end
	for i = 0, Map.GetNumPlots() - 1 do
		terrainTypes[i] = defaultTerrain
	end
	SetTerrainTypes(terrainTypes);
	
	--key variables in terrain generation
	local grassPercent
	local plainsPercent
	local tundraPercent
	local wastesPercent
	local grassPercentVariation
	local plainsPercentVariation
	local tundraPercentVariation
	local wastesPercentVariation
	if defaultTerrain == TerrainTypes.TERRAIN_SNOW then
		grassPercent = 35
		plainsPercent = 30
		tundraPercent = 20
		wastesPercent = 15
		grassPercentVariation = 10
		plainsPercentVariation = 5
		tundraPercentVariation = 4
		wastesPercentVariation = 4
	else
		grassPercent = 40
		plainsPercent = 38
		tundraPercent = 0
		wastesPercent = 22
		grassPercentVariation = 10
		plainsPercentVariation = 10
		tundraPercentVariation = 0
		wastesPercentVariation = 5
	end
	
	
	local minMask = .25
	local maxMask = 1.0
	local fractalLayers = 3 --more layers create less noise
	
	for i = 1, table.getn(starting_points) do
		local start_size;
		if starting_points[i].cityState == false then
			start_size = player_size
		else
			start_size = cs_size
		end
		
		--apply variation to terrain percentages
		grassPercentPlayer = grassPercent + (GetRandom(0, grassPercentVariation * 2) - grassPercentVariation)
		plainsPercentPlayer = plainsPercent + (GetRandom(0, plainsPercentVariation * 2) - plainsPercentVariation)
		tundraPercentPlayer = tundraPercent + (GetRandom(0, tundraPercentVariation * 2) - tundraPercentVariation)
		wastesPercentPlayer = wastesPercent + (GetRandom(0, wastesPercentVariation * 2) - wastesPercentVariation)
		local normalizePercent = 100 / (wastesPercentPlayer + tundraPercentPlayer + plainsPercentPlayer + grassPercentPlayer)
		grassPercentPlayer = Round(grassPercentPlayer * normalizePercent)
		plainsPercentPlayer = Round(plainsPercentPlayer * normalizePercent)
		tundraPercentPlayer = Round(tundraPercentPlayer * normalizePercent)
		wastesPercentPlayer = Round(wastesPercentPlayer * normalizePercent)
		
		local grassPercentile = wastesPercentPlayer + tundraPercentPlayer + plainsPercentPlayer
		local plainsPercentile = wastesPercentPlayer + tundraPercentPlayer
		local tundraPercentile = wastesPercentPlayer
		
		--init terrain mask and final terrain tables
		local terrainMask = {}
		local terrain = {}
		
		--generate terrain mask for this size
		for j = 1, table.getn(starting_points[i].plots) do
			local localX = starting_points[i].plots[j].localX
			local localY = starting_points[i].plots[j].localY
			local distance = starting_points[i].plots[j].depth
			
			--these tables are always accessed sequentially, so we don't need them to be two-dimensional
			table.insert(terrainMask, {
				x = localX, 
				y = localY, 
				value = maxMask - ((distance) * ((maxMask - minMask) / (start_size)))
			})
		end
		
		--generate x fractals, apply mask, and then average them into the final terrain
		for j = 1, fractalLayers do
			local playerTerrain = Fractal.Create((start_size * 2) + 1, (start_size * 2) + 1, 4, {}, 4, 4)
			
			for k = 1, table.getn(terrainMask) do
				local x = terrainMask[k].x
				local y = terrainMask[k].y
				
				local height = playerTerrain:GetHeight(x, y) * terrainMask[k].value
				if j == 1 then
					table.insert(terrain, height)
				else
					terrain[k] = (terrain[k] + height) / 2
				end
			end
		end
		
		--find max val for normalize
		local maxVal = 0
		for j = 1, table.getn(terrain) do
			if terrain[j] > maxVal then maxVal = terrain[j] end
		end
		
		--normalize terrrain
		for j = 1, table.getn(terrain) do
			terrain[j] = 255 / maxVal * terrain[j]
		end
		
		--sort values for percentiles. have to copy first since table assignments are by reference
		local sortedTerrain = {}
		for j = 1, table.getn(terrain) do
			sortedTerrain[j] = terrain[j]
		end
		table.sort(sortedTerrain)
		
		--find percentiles
		local grassThreshold = Round(sortedTerrain[math.ceil(table.getn(sortedTerrain) * (grassPercentile / 100))])
		local plainsThreshold = Round(sortedTerrain[math.ceil(table.getn(sortedTerrain) * (plainsPercentile / 100))])
		local tundraThreshold = Round(sortedTerrain[math.ceil(table.getn(sortedTerrain) * (tundraPercentile / 100))])
		--[[
		print ("Player ".. i .. " > " .. grassPercentPlayer .. "% Grass (" .. grassPercentile .."th): " .. grassThreshold)
		print ("Player ".. i .. " > " .. plainsPercentPlayer .. "% Plains (" .. plainsPercentile .."th): " .. plainsThreshold)
		print ("Player ".. i .. " > " .. tundraPercentPlayer .. "% Tundra (" .. tundraPercentile .."th): " .. tundraThreshold)
		print ("Player ".. i .. " > " .. wastesPercentPlayer .. "% Wastes")
		]]
		
		--create terrain
		for j = 1, table.getn(starting_points[i].plots) do
			local x = starting_points[i].plots[j].x
			local y = starting_points[i].plots[j].y
			local localX = starting_points[i].plots[j].localX
			local localY = starting_points[i].plots[j].localY
			
			local height = terrain[j]
			
			--if i == 1 then print(j .. ": " .. height) end

			local plot = Map.GetPlot(x, y)
			--print (x .. ", " .. y .. ": " .. height)
			if height > grassThreshold then
				plot:SetTerrainType(TerrainTypes.TERRAIN_GRASS, false, false);
			elseif height > plainsThreshold then
				plot:SetTerrainType(TerrainTypes.TERRAIN_PLAINS, false, false);
			elseif height > tundraThreshold then
				if defaultTerrain == TerrainTypes.TERRAIN_SNOW then
					plot:SetTerrainType(TerrainTypes.TERRAIN_TUNDRA, false, false);
				else
					plot:SetTerrainType(TerrainTypes.TERRAIN_PLAINS, false, false);
				end
			else
				plot:SetTerrainType(defaultTerrain, false, false);
			end
		end
	end
end

lakes = {}
function AddLakes()
	--each starting point has a 50% chance to have a lake
	local lakeSeed = {}
	for i = 1, table.getn(starting_points) do
		local lakeChance
		if starting_points[i].cityState == false then
			lakeChance = 5
		else
			lakeChance = 8
		end
	
		if GetRandom(1, 10) > lakeChance then
			print("Player " .. i .. " has a lake.")
			
			--find the seed location of the lake within 0 to 2 tiles of the start pos
			local seedDir = Map.Rand(6, "")
			local seedDist = Map.Rand(2, "")
			--print("Dist: " .. seedDist .. ", Dir: " .. seedDir)
			if seedDist == 0 then
				--move one plot away
				local adjPlot = Map.PlotDirection(starting_points[i].x, starting_points[i].y, seedDir)
				table.insert(lakeSeed, {x = adjPlot:GetX(), y = adjPlot:GetY(), cityState = starting_points[i].cityState})
			else
				--move one plot away
				local adjPlot = Map.PlotDirection(starting_points[i].x, starting_points[i].y, seedDir)
				
				--random adjustment to direction -1 to +1
				seedDir = (seedDir + Map.Rand(3, "") - 1) % 6; -- mod 6 to take care of wrapping
				
				--move two plots away
				adjPlot = Map.PlotDirection(adjPlot:GetX(), adjPlot:GetY(), seedDir)
				table.insert(lakeSeed, {x = adjPlot:GetX(), y = adjPlot:GetY(), cityState = starting_points[i].cityState})
			end
		end
	end
	
	for i = 1, table.getn(lakeSeed) do
		--give the lake a random size between 1 and 4 plots
		local lakeSize
		if lakeSeed[i].cityState == true then
			lakeSize = Map.Rand(3, "") + 1
		else
			lakeSize = Map.Rand(4, "") + 1
		end
		
		local plot = Map.GetPlot(lakeSeed[i].x, lakeSeed[i].y)
		local lakePlots = {}
		
		for j = 1, lakeSize do
			plot:SetPlotType(PlotTypes.PLOT_OCEAN, false, false)
			table.insert(lakePlots, {x = plot:GetX(), y = plot:GetY()})
			
			if j ~= lakeSize then
				--now choose a random direction to spread the lake
				local spreadDir = Map.Rand(6, "")
				local k = 0
				local nextPlot
				repeat
					nextPlot = Map.PlotDirection(plot:GetX(), plot:GetY(), (spreadDir + k) % 6)
					if nextPlot ~= nil then --possible map boundary check
						if nextPlot:GetPlotType() ~= PlotTypes.PLOT_OCEAN and nextPlot:GetTerrainType() ~= TerrainTypes.TERRAIN_SNOW and nextPlot:GetTerrainType() ~= TerrainTypes.TERRAIN_DESERT then
							--this plot is acceptable, so break out of loop
							break
						end
					end
					k = k + 1
				until k > 5
				plot = nextPlot
			end
		end
		
		table.insert(lakes, lakePlots)
	end
end

function AddRivers()
	for i = 1, table.getn(lakes) do
		local riverChance = Map.Rand(100, "")
		local numRivers
		
		--chance to have river
		if riverChance < 80 then
			--chance to have two rivers; increases the larger the lake is
			if riverChance < 5 * table.getn(lakes[i]) then
				numRivers = 3
			elseif riverChance < 12 * table.getn(lakes[i]) then
				numRivers = 2
			else
				numRivers = 1
			end
			
			for j = 1, numRivers do
				local riverPlot = GetRandom(1, table.getn(lakes[i]))
				local riverLength = GetRandom(1, player_size + 1)
				SetupRiver(lakes[i][riverPlot].x, lakes[i][riverPlot].y, riverLength)
			end
		end
	end
end

local rivers = {}
function SetupRiver(x, y, length)
	print ("River at " .. x .. ", " .. y .. ", Length:" .. length)
	local riverPlot = Map.GetPlot(x, y)
	
	table.insert(rivers, {
		startPlot = riverPlot, 
		forks = 0,
		lengthLeft = length
	})
	local this_river = #rivers
	
	--find the direction of the nearest start plot
	local nearest_start_dir
	for i = 1, table.getn(starting_points) do
		if (Map.PlotDistance(starting_points[i].x, starting_points[i].y, x, y) < 5) then
			--rough estimate of direction. hex math is weird.
			local xDiff = starting_points[i].x - x
			local yDiff = starting_points[i].y - y
			
			if yDiff < 0 then
				if xDiff < 0 then
					nearest_start_dir = DirectionTypes.DIRECTION_SOUTHWEST
				else
					nearest_start_dir = DirectionTypes.DIRECTION_SOUTHEAST
				end
			elseif yDiff > 0 then
				if xDiff < 0 then
					nearest_start_dir = DirectionTypes.DIRECTION_NORTHWEST
				else
					nearest_start_dir = DirectionTypes.DIRECTION_NORTHEAST
				end
			else --x diff == 0
				if xDiff < 0 then
					nearest_start_dir = DirectionTypes.DIRECTION_WEST
				else
					nearest_start_dir = DirectionTypes.DIRECTION_EAST
				end
			end
		end
	end
	
	--pick a random direction based on suitableness of the surrounding plots
	local adjPlots = {}
	for i = 0, 5 do
		table.insert(adjPlots, {suitableness = EvalRiverPlot(Map.PlotDirection(x, y, i)), direction = i})
		--add in the nearest start direction bonus
		if adjPlots[#adjPlots].suitableness ~= nil and nearest_start_dir ~= nil then
			if adjPlots[#adjPlots].suitableness > 0 then
				local relative_to_start = math.abs(i - nearest_start_dir)
				if relative_to_start == 0 then
					adjPlots[#adjPlots].suitableness = adjPlots[#adjPlots].suitableness * 2
				end
			end
		end
	end
	
	--now work out direction suitableness based on the two adjacent plots
	local flowDirections = {}
	local directionSuitable
	local totalSuitable = 0
	-- adjPlots 1 to 6
	for i = 0, 5 do
		local left
		local right
		if i < 1 then
			left = adjPlots[6]; --wrap
		else
			left = adjPlots[i]
		end
		if i > 5 then
			right = adjPlots[1]; --wrap
		else
			right = adjPlots[i + 1]
		end
		
		--make sure neither is a lake or the edge of the map
		if left.suitableness ~= nil and right.suitableness ~= nil then
			if left.suitableness > 0 and right.suitableness > 0 then
				directionSuitable = left.suitableness + right.suitableness
				totalSuitable = totalSuitable + directionSuitable
				--print("Flow Dir: " .. i .. ", Suitable: " .. directionSuitable)
				table.insert(flowDirections, {suitableness = directionSuitable, flowDirection = i})
			end
		end
	end
	
	
	--pick a random flow direction based on suitableness
	local flowDirection
	local randSuitable = GetRandom(1, totalSuitable)
	if table.getn(flowDirections) > 0 then
		for i = 1, table.getn(flowDirections) do
			totalSuitable = totalSuitable - flowDirections[i].suitableness
			if randSuitable > totalSuitable then
				flowDirection = flowDirections[i].flowDirection
				--print("River Direction: " .. flowDirection)
				break
			end
		end
	else
		return --no suitable directions
	end
	
	--now we know where we want the river to go, so put it there
	--print (x .. ", " .. y .. " | " ..  flowDirection)
	if flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH or flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTH then
		if flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_NORTHWEST)
		else
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_SOUTHWEST)
		end
		
		if riverPlot == nil then return nil end
		riverPlot:SetWOfRiver(true, GetOppositeFlowDirection(flowDirection))
	elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST or flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST then
		if flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_NORTHEAST)
		else
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_WEST)
		end
		
		if riverPlot == nil then return nil end
		riverPlot:SetNWOfRiver(true, GetOppositeFlowDirection(flowDirection))
	elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST or flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST then
		if flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_EAST)
		else 
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_NORTHWEST)
		end
		
		if riverPlot == nil then return nil end
		riverPlot:SetNEOfRiver(true, GetOppositeFlowDirection(flowDirection))
	end
	
	--now generate the river out from here
	rivers[this_river].generalDirection = flowDirection
	RiverFlow(riverPlot, this_river, flowDirection)
end
	
function RiverFlow(riverPlot, this_river, flowDirection)
	--loop though the length
	local previousDirection
	local length = rivers[this_river].lengthLeft
	for i = 1, length do
		--there are only two directions that the river can go now
		--so we only have to evaluate 3 plots for their suitableness
		local adjPlots = {}
		previousDirection = flowDirection
		if flowDirection >= 0 and flowDirection <= 2 then
			adjPlots = {
				left = {plot = riverPlot},
				center = {plot = Map.PlotDirection(riverPlot:GetX(), riverPlot:GetY(), flowDirection)},
				right = {plot = Map.PlotDirection(riverPlot:GetX(), riverPlot:GetY(), (flowDirection + 1) % 6)},
			}
		else
			--left and right are reversed for this direction
			adjPlots = {
				left = {plot = Map.PlotDirection(riverPlot:GetX(), riverPlot:GetY(), (flowDirection - 2) % 6)},
				center = {plot = Map.PlotDirection(riverPlot:GetX(), riverPlot:GetY(), (flowDirection - 1) % 6)},
				right = {plot = riverPlot},
			}
		end
		
		--don't go off the edge of the map
		if adjPlots.center.plot == nil then return end
		
		--calculate suitableness
		adjPlots.left.suitable = EvalRiverPlot(adjPlots.left.plot)
		adjPlots.center.suitable = EvalRiverPlot(adjPlots.center.plot)
		adjPlots.right.suitable = EvalRiverPlot(adjPlots.right.plot)
		
		if adjPlots.center.plot:GetTerrainType() == TerrainTypes.TERRAIN_SNOW or adjPlots.center.plot:GetPlotType() == PlotTypes.PLOT_OCEAN then
			--don't let the rivers flow into snow or back into a lake
			--we're allowing desert rivers (for now, at least) since they're less absurd than rivers in snow
			return
		end
		
		local leftSuitable = adjPlots.left.suitable + adjPlots.center.suitable
		local rightSuitable = adjPlots.center.suitable + adjPlots.right.suitable
		local totalSuitable = leftSuitable + rightSuitable
		local randSuitable = Map.Rand(totalSuitable, "")
		local randFork = Map.Rand(100, "")
		local forkChance = (rivers[this_river].lengthLeft * 5) - (rivers[this_river].forks * 10)
		local leftDirection = (flowDirection - 1) % 6
		local rightDirection = (flowDirection + 1) % 6
		
		--apply general direction boundary
		if leftDirection == GetOppositeFlowDirection(rivers[this_river].generalDirection) then
			leftDirection = rightDirection
			forkChance = 0
		elseif rightDirection == GetOppositeFlowDirection(rivers[this_river].generalDirection) then
			rightDirection = leftDirection
			forkChance = 0
		end
		
		if totalSuitable >= 200 and randFork < forkChance then
			--chance to fork the river
			
			local forkPlot = AddRiverFromPlot(riverPlot, leftDirection, previousDirection)
			if forkPlot ~= nil then
				--print("fork {")
				rivers[this_river].forks = rivers[this_river].forks + 1
				RiverFlow(forkPlot, this_river, leftDirection)
				--print("}")
				flowDirection = rightDirection
			end
		else
			if leftSuitable > randSuitable then
				flowDirection = leftDirection
			else
				flowDirection = rightDirection
			end
		end
		
		riverPlot = AddRiverFromPlot(riverPlot, flowDirection, previousDirection)
		
		if riverPlot == nil then break end
		
		rivers[this_river].lengthLeft = rivers[this_river].lengthLeft - 1
		--print ("  length left: " .. rivers[this_river].lengthLeft)
	end
end

function AddRiverFromPlot(riverPlot, flowDirection, previousDirection)
--[[
"DIRECTION_NORTHEAST" 		0
"DIRECTION_EAST" 			1
"DIRECTION_SOUTHEAST" 		2
"DIRECTION_SOUTHWEST" 		3
"DIRECTION_WEST" 			4
"DIRECTION_NORTHWEST" 		5 
	
"FLOWDIRECTION_NORTH" 		0
"FLOWDIRECTION_NORTHEAST" 	1
"FLOWDIRECTION_SOUTHEAST" 	2
"FLOWDIRECTION_SOUTH" 		3
"FLOWDIRECTION_SOUTHWEST" 	4
"FLOWDIRECTION_NORTHWEST" 	5 

don't change POS
	clockwise
		s - sw
		sw - nw
	counterclockwise
		se - ne
		ne - n
change POS
	clockwise
		nw - n
		n - ne
		ne - se
		se - s
	counterclockwise
		n - nw
		nw - sw
		sw - s
		s - se
]]
	local x = riverPlot:GetX()
	local y = riverPlot:GetY()

	if flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH or flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTH then
		--whether or not to move plots depends on the direction of the previous flow direction and the intended direction
		if flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH and previousDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_WEST)
		elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTH then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_SOUTHWEST)
		end
		
		if riverPlot == nil then return nil end
		
		--must turn the direction of flow around because the direction of generation is opposite of flow
		--if we don't do this, our rivers will look silly
		riverPlot:SetWOfRiver(true, GetOppositeFlowDirection(flowDirection))
	elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST or flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST then
		if flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST and previousDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_NORTHEAST)
		elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST and previousDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_WEST)
		end
		
		if riverPlot == nil then return nil end
		riverPlot:SetNWOfRiver(true, GetOppositeFlowDirection(flowDirection))
	elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST or flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST then
		if flowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_EAST)
		elseif flowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST and previousDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH then
			riverPlot = Map.PlotDirection(x, y, DirectionTypes.DIRECTION_NORTHEAST)
		end
		
		if riverPlot == nil then return nil end
		riverPlot:SetNEOfRiver(true, GetOppositeFlowDirection(flowDirection))
	end
	
	--print (x .. ", " .. y .. " | " ..  flowDirection .. ", " .. previousDirection)
	
	return riverPlot
end

function EvalRiverPlot(thisPlot)
	local plotSuitable = 0
	
	if thisPlot == nil then return 0 end --boundary hit
	
	--judge the suitableness of plot types
	if thisPlot:GetPlotType() == PlotTypes.PLOT_OCEAN then
		--lake tiles are completely unsuitable
		plotSuitable = 0
	elseif thisPlot:GetPlotType() == PlotTypes.PLOT_LAND then
		plotSuitable = 45
	elseif thisPlot:GetPlotType() == PlotTypes.PLOT_HILLS then
		plotSuitable = 30
	elseif thisPlot:GetPlotType() == PlotTypes.PLOT_MOUNTAIN then
		plotSuitable = 10
	end
	
	if plotSuitable > 0 then
		--judge the suitableness of terrain types
		if thisPlot:GetTerrainType() == TerrainTypes.TERRAIN_GRASS then
			plotSuitable = plotSuitable + 60
		elseif thisPlot:GetTerrainType() == TerrainTypes.TERRAIN_PLAINS then
			plotSuitable = plotSuitable + 30
		elseif thisPlot:GetTerrainType() == TerrainTypes.TERRAIN_TUNDRA then
			plotSuitable = plotSuitable + 10
		elseif thisPlot:GetTerrainType() == TerrainTypes.TERRAIN_DESERT then
			plotSuitable = plotSuitable - 30
		elseif thisPlot:GetTerrainType() == TerrainTypes.TERRAIN_SNOW then
			--don't go into snow either
			plotSuitable = 0
		end
	end
	
	return plotSuitable
end

function GetOppositeFlowDirection(direction)
	local numTypes = FlowDirectionTypes.NUM_FLOWDIRECTION_TYPES;
	return ((direction + 3) % numTypes);
end

function AddFeatures()
	--generate jungle and forest for each player start
	for i = 1, table.getn(starting_points) do
		local start_size, jungleThreshold, forestThreshold
		if starting_points[i].cityState == false then
			start_size = player_size
		else
			start_size = cs_size
		end
		
		local playerFlora = Fractal.Create((start_size * 2) + 1, (start_size * 2) + 1, 3, {}, -1, -1)
		
		if start_size == 5 then
			jungleThreshold = {
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(95),
				256
			}
			forestThreshold = {
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(65),
				playerFlora:GetHeight(60),
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(90)
			}
		elseif start_size == 4 then
			jungleThreshold = {
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(95),
				256
			}
			forestThreshold = {
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(60),
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(90)
			}
		elseif start_size == 3 then
			jungleThreshold = {
				playerFlora:GetHeight(90),
				playerFlora:GetHeight(60),
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(90)
			}
			forestThreshold = {
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(50),
				playerFlora:GetHeight(50),
				playerFlora:GetHeight(70)
			}
		else --if start_size == 2 then
			jungleThreshold = {
				playerFlora:GetHeight(80),
				playerFlora:GetHeight(60),
				playerFlora:GetHeight(90)
			}
			forestThreshold = {
				playerFlora:GetHeight(70),
				playerFlora:GetHeight(50),
				playerFlora:GetHeight(60)
			}
		end
		
		for j = 1, table.getn(starting_points[i].plots) do
			local x = starting_points[i].plots[j].x
			local y = starting_points[i].plots[j].y
			local localX = starting_points[i].plots[j].localX
			local localY = starting_points[i].plots[j].localY
			local depth = starting_points[i].plots[j].depth
			
			local height = playerFlora:GetHeight(localX, localY)
			local plot = Map.GetPlot(x, y)
			
			if height >= jungleThreshold[depth + 1] and plot:GetTerrainType() ~= TerrainTypes.TERRAIN_SNOW and plot:GetTerrainType() ~= TerrainTypes.TERRAIN_TUNDRA and plot:GetTerrainType() ~= TerrainTypes.TERRAIN_DESERT and plot:GetPlotType() ~= PlotTypes.PLOT_OCEAN then
				plot:SetFeatureType(FeatureTypes.FEATURE_JUNGLE, -1)
			elseif height >= forestThreshold[depth + 1] and plot:GetTerrainType() ~= TerrainTypes.TERRAIN_SNOW and plot:GetTerrainType() ~= TerrainTypes.TERRAIN_DESERT and plot:GetPlotType() ~= PlotTypes.PLOT_OCEAN then
				plot:SetFeatureType(FeatureTypes.FEATURE_FOREST, -1)
			end
		end
	end
	
	--generate marshes for players
	for i = 1, table.getn(starting_points) do
		local start_size
		if starting_points[i].cityState == false then
			start_size = player_size
		else
			start_size = cs_size
		end
		
		local playerMarsh = Fractal.Create((start_size * 2) + 1, (start_size * 2) + 1, 3, {}, -1, -1)
		local marshThreshold = playerMarsh:GetHeight(80)
		local marshCount = 0
		
		for j = 1, table.getn(starting_points[i].plots) do
			local x = starting_points[i].plots[j].x
			local y = starting_points[i].plots[j].y
			local localX = starting_points[i].plots[j].localX
			local localY = starting_points[i].plots[j].localY
			
			local height = playerMarsh:GetHeight(localX, localY)
			local plot = Map.GetPlot(x, y)
			
			--must be:
			--	grassland
			--	flat land
			--  no other features
			--	river or lake adjacent
			if height >= marshThreshold and plot:GetTerrainType() == TerrainTypes.TERRAIN_GRASS and plot:GetPlotType() == PlotTypes.PLOT_LAND and plot:GetFeatureType() == FeatureTypes.NO_FEATURE and (plot:IsCoastalLand() or plot:IsRiverSide()) then
				plot:SetFeatureType(FeatureTypes.FEATURE_MARSH, -1)
				marshCount = marshCount + 1
			end
			
			--limit amount of marsh because it's annoying
			if marshCount > start_size then break end
		end
	end
end

--[[
--disabling this for now. new terrain generation code creates far less holes (with current settings)
--and when it does create holes, they look fairly natural
function FixTerrainHoles()
	--remove terrain holes
	for i = 1, table.getn(starting_points) do
		for j = 1, table.getn(starting_points[i].plots) do
			local x = starting_points[i].plots[j].x
			local y = starting_points[i].plots[j].y
			local localX = starting_points[i].plots[j].localX
			local localY = starting_points[i].plots[j].localY
			
			local plot = Map.GetPlot(x, y)
			local plotTerrain = plot:GetTerrainType() + 1
			
			local adjTerrains = table.fill(0, TerrainTypes.NUM_TERRAIN_TYPES)
			for k = 0, 5 do
				local adjPlot = Map.PlotDirection(x, y, k)
				if adjPlot ~= nil then --map boundaries check
					adjTerrain = adjPlot:GetTerrainType() + 1
					adjTerrains[adjTerrain] = adjTerrains[adjTerrain] + 1
				end
			end
			
			--see if there are any adjacent plots with the same terrain
			if adjTerrains[plotTerrain] < 1 then
				--print ("Terrain hole at " .. x ..  ", " .. y)
				
				--chance to just leave the hole alone and make it look a little quirky
				if Map.Rand(10, "") >= -1 then
					--more checks
					local holeOK = false
					if plotTerrain >= TerrainTypes.TERRAIN_COAST + 1 then
						--coast, hills, and mountains shouldn't be included
						holeOK = true
					--elseif plotTerrain == TerrainTypes.TERRAIN_SNOW + 1 and adjTerrains[TerrainTypes.TERRAIN_TUNDRA + 1] > 0 then
						--snow next to tundra is okay
						--holeOK = true
					elseif plotTerrain == TerrainTypes.TERRAIN_TUNDRA + 1 and adjTerrains[TerrainTypes.TERRAIN_SNOW + 1] > 0 then
						--tundra next to snow is okay
						holeOK = true
					--elseif plotTerrain == TerrainTypes.TERRAIN_TUNDRA + 1 and adjTerrains[TerrainTypes.TERRAIN_PLAINS + 1] > 0 then
						--tundra next to plains is okay
						--holeOK = true
					elseif plotTerrain == TerrainTypes.TERRAIN_PLAINS + 1 and adjTerrains[TerrainTypes.TERRAIN_GRASS + 1] > 0 then
						--plains next to grassland is okay
						holeOK = true
					elseif plotTerrain == TerrainTypes.TERRAIN_GRASS + 1 and adjTerrains[TerrainTypes.TERRAIN_PLAINS + 1] > 0 then
						--grassland next to plains is okay
						holeOK = true
					elseif plotTerrain == TerrainTypes.TERRAIN_PLAINS + 1 and adjTerrains[TerrainTypes.TERRAIN_DESERT + 1] > 0 then
						--plains next to desert is okay
						holeOK = true
					end
					
					if holeOK == false then
						--pick an adjacent terrain and switch this plot to that
						local adjPlot = Map.PlotDirection(x, y, Map.Rand(6, "")) 
						if adjPlot ~= nil then
							--print("setting " .. plot:GetTerrainType() .. " to " .. adjPlot:GetTerrainType())
							plot:SetTerrainType(adjPlot:GetTerrainType(), false, false)
						end
					end
				end
			end
		end
	end
end
]]

function PlaceStartPlots()
	--move human player to most fertile start for epic start
	if epic_start >= 4 then
		--[[
		local mostAwesomeStart = 1
		local mostAwesomeAmt = 0
		for i = 1, table.getn(starting_points) do
			if starting_points[i].cityState == false then
				local awesomeTotal = 0
				for j = 1, table.getn(starting_points[i].plots) do
					awesomeTotal = awesomeTotal + GetPlotAwesomeness(Map.GetPlot(starting_points[i].plots[j].x, starting_points[i].plots[j].y))
				end
				if awesomeTotal > mostAwesomeAmt then
					mostAwesomeAmt = awesomeTotal
					mostAwesomeStart = i
				end
				print("Player " .. i .. " Start Awesomeness: " .. awesomeTotal)
			end
		end
		print("Most Awesome Start (Old Method): " .. mostAwesomeStart)
		]]
		
		local mostAwesomeStart = 1
		local mostAwesomeAmt = 0
		
		local bestHillRatio = .35
		local bestGrassRatio = .65
		local bestFloraRatio = .4
		local bestFreshWaterRatio = .45
		
		for i = 1, table.getn(starting_points) do
			if starting_points[i].cityState == false then
				local awesomeTotal = 0
				
				local centerPlots = GetPlotsAtDistance(starting_points[i].x, starting_points[i].y, 0, 3)
				local edgePlots
				if player_size > 3 then
					edgePlots = GetPlotsAtDistance(starting_points[i].x, starting_points[i].y, 4, player_size)
				end
				local plotAreas = {centerPlots, edgePlots}
				
				for j = 1, table.getn(plotAreas) do
					if plotAreas[j] ~= nil then
						local hillCount = 0
						local grassCount = 0
						local floraCount = 0
						local freshWaterCount = 0
						local hillRatio, grassRatio, floraRatio, freshWaterRatio, hillRatioDiff, grassRatioDiff, floraRatioDiff, freshWaterRatioDiff
						for k = 1, table.getn(plotAreas[j]) do
							local plot = Map.GetPlot(plotAreas[j][k].x, plotAreas[j][k].y)
							if plot:GetPlotType() == PlotTypes.PLOT_HILLS then
								hillCount = hillCount + 1
							end
							if plot:GetTerrainType() == TerrainTypes.TERRAIN_GRASS and plot:GetPlotType() == PlotTypes.PLOT_LAND then
								grassCount = grassCount + 1
							end
							if plot:GetFeatureType() == FeatureTypes.FEATURE_JUNGLE or plot:GetFeatureType() == FeatureTypes.FEATURE_FOREST then
								floraCount = floraCount + 1
							end
							if plot:IsCoastalLand() or plot:IsRiverSide() then
								freshWaterCount = freshWaterCount + 1
							end
						end
						
						hillRatio = hillCount / table.getn(plotAreas[j])
						grassRatio = grassCount / table.getn(plotAreas[j])
						floraRatio = floraCount / table.getn(plotAreas[j])
						freshWaterRatio = freshWaterCount / table.getn(plotAreas[j])
						
						hillRatioDiff = math.abs(hillRatio - bestHillRatio) * (1 / bestHillRatio)
						grassRatioDiff = math.abs(grassRatio - bestGrassRatio) * (1 / bestGrassRatio)
						floraRatioDiff = math.abs(floraRatio - bestFloraRatio) * (1 / bestFloraRatio)
						freshWaterRatioDiff = math.abs(freshWaterRatio - bestFreshWaterRatio) * (1 / bestFreshWaterRatio)
						
						--multiplier is different for edge plots and center plots
						local ratioMult = 50
						if j == 2 then ratioMult = 20 end --edge plots
						awesomeTotal = awesomeTotal + Round(
							(math.abs(hillRatioDiff - 1) * ratioMult) + 
							(math.abs(grassRatioDiff - 1) * ratioMult) + 
							(math.abs(floraRatioDiff - 1) * ratioMult) + 
							(math.abs(freshWaterRatioDiff - 1) * ratioMult)
						)
						
						
						--print(i .. " > hillRatio: " .. Round(hillRatio * 100) .. "%, grassRatio: " .. Round(grassRatio * 100) .. "% floraRatio: " .. Round(floraRatio * 100) .. "% freshWaterRatio: " .. Round(freshWaterRatio * 100) .. "%")
						
					end
				end
				
				if awesomeTotal > mostAwesomeAmt then
					mostAwesomeAmt = awesomeTotal
					mostAwesomeStart = i
				end
				print("Player " .. i .. " Start Awesomeness: " .. awesomeTotal)
			end
		end
		print("Most Awesome Start (New Method): " .. mostAwesomeStart)
		
		
		if mostAwesomeStart ~= 1 then
			--if player is not already at best start then flip them
			local tempStart = starting_points[1]
			starting_points[1] = starting_points[mostAwesomeStart]
			starting_points[mostAwesomeStart] = tempStart
		end
	end

	local cs = 1
	for i = 1, table.getn(starting_points) do
		local player
		if starting_points[i].cityState == false then
			player = Players[i - 1]
		else
			--city-states start > i = 1
			player = Players[cs + (GameDefines.MAX_MAJOR_CIVS - 1)]
			cs = cs + 1
		end
		
		local start_plot = Map.GetPlot(starting_points[i].x, starting_points[i].y)
		--if the current start plot is a mountain or water, then randomly pick another nearby tile that is clear
		if start_plot:GetPlotType() == PlotTypes.PLOT_MOUNTAIN or start_plot:GetPlotType() == PlotTypes.PLOT_OCEAN then
			print("player " .. i .. " start blocked")
			local j = 0
			local randDir = GetRandom(0, 5)
			repeat
				randDir = (randDir + j) % 6
				start_plot = Map.PlotDirection(starting_points[i].x, starting_points[i].y, randDir)
				if start_plot ~= nil then
					if start_plot:GetPlotType() ~= PlotTypes.PLOT_MOUNTAIN and start_plot:GetPlotType() ~= PlotTypes.PLOT_OCEAN then
						break
					end
				end
				j = j + 1
			until j > 6
			print("new start: " .. randDir)
		end
		start_plot = start_plot:GetNearestLandPlot()
		player:SetStartingPlot(start_plot)
	end
end

--[[
function GetPlotAwesomeness(plot)
	local this_plot_type = plot:GetPlotType()
	local this_plot_terrain = plot:GetTerrainType()
	local this_plot_feature = plot:GetFeatureType()
	
	local awesomeness = 0
	
	--plot type
	if this_plot_type == PlotTypes.PLOT_LAND then
		awesomeness = awesomeness + 0 --flat land depends on what's on it
	elseif this_plot_type == PlotTypes.PLOT_HILLS then
		awesomeness = awesomeness + 3 --hills are great all on their own
	elseif this_plot_type == PlotTypes.PLOT_MOUNTAIN then
		awesomeness = awesomeness + 1
	elseif this_plot_type == PlotTypes.PLOT_OCEAN then
		awesomeness = awesomeness + 1 --lakes aren't that great in their own right
	end
	--terrain
	if this_plot_terrain == TerrainTypes.TERRAIN_GRASS then
		awesomeness = awesomeness + 3
	elseif this_plot_terrain == TerrainTypes.TERRAIN_PLAINS then
		awesomeness = awesomeness + 2
	elseif this_plot_terrain == TerrainTypes.TERRAIN_TUNDRA then
		awesomeness = awesomeness + 1
	elseif this_plot_terrain == TerrainTypes.TERRAIN_SNOW or this_plot_terrain == TerrainTypes.TERRAIN_DESERT then
		awesomeness = awesomeness - 2 --can't be worked unless they're hills
	end
	--features
	if this_plot_feature == FeatureTypes.FEATURE_JUNGLE then
		awesomeness = awesomeness + 1
	elseif this_plot_feature == FeatureTypes.FEATURE_FOREST then
		awesomeness = awesomeness + 1
	elseif this_plot_feature == FeatureTypes.FEATURE_MARSH then
		awesomeness = awesomeness - 1 --marshes suck but if you remove them, they aren't that bad
	end
	--fresh water
	if plot:IsCoastalLand() or plot:IsRiverSide() then
		awesomeness = awesomeness + 2
	end
	
	return awesomeness
end
]]

function PlaceNaturalWonders()
	--my version of this isn't as detailed as AssignStartedPlots.lua but allows more control and i like that

	abundance_values = {Sparse = 1, Normal = 2, Abundant = 3}
	resource_abundance = Map.GetCustomOption(5)

	--first setup wonders
	local natural_wonders = {}
	
	if Map.GetCustomOption(7) <= 5 then
		--dongs already have geysers
		table.insert(natural_wonders, {name = "geyser", id = GameInfo.Features.FEATURE_GEYSER.ID}) --geyser go anywhere but desert. change to mountain
	end
	table.insert(natural_wonders, {name = "crater", id = GameInfo.Features.FEATURE_CRATER.ID}) --can be on desert or tundra. change to mountain
	table.insert(natural_wonders, {name = "mtfuji", id = GameInfo.Features.FEATURE_FUJI.ID}) --can be on grass or plains. change to mountain
	table.insert(natural_wonders, {name = "mesa", id = GameInfo.Features.FEATURE_MESA.ID}) --plains, desert, or tundra. change to mountain
	table.insert(natural_wonders, {name = "fountain", id = GameInfo.Features.FEATURE_FOUNTAIN_YOUTH.ID}) --grassland or plains. change to flat land
	table.insert(natural_wonders, {name = "potosi", id = GameInfo.Features.FEATURE_POTOSI.ID}) --can be anywhere. change to mountain
	table.insert(natural_wonders, {name = "eldorado", id = GameInfo.Features.FEATURE_EL_DORADO.ID}) --grassland or plains. change to flat land
	
	if isGAndK == true then
		table.insert(natural_wonders, {name = "sripada", id = GameInfo.Features.FEATURE_SRI_PADA.ID}) --grassland or plains. change to mountain
		table.insert(natural_wonders, {name = "mtsinai", id = GameInfo.Features.FEATURE_MT_SINAI.ID}) --plains or desert. change to mountain
		table.insert(natural_wonders, {name = "mtkailash", id = GameInfo.Features.FEATURE_MT_KAILASH.ID}) --grassland or plains. change to mountain
		table.insert(natural_wonders, {name = "uluru", id = GameInfo.Features.FEATURE_ULURU.ID}) --desert or plains. change to mountain
	end
	
	--decide on how many to place given map size and also resources abundance setting
	local sizevalues = {
		[0] = 1, --WORLDSIZE_DUEL
		[1] = 2, --WORLDSIZE_TINY
		[2] = 3, --WORLDSIZE_SMALL
		[3] = 4, --WORLDSIZE_STANDARD
		[4] = 6, --WORLDSIZE_LARGE
		[5] = 8 --WORLDSIZE_HUGE
	};
	local numWonders = sizevalues[Map.GetWorldSize()] or 1
	
	if resource_abundance == abundance_values.Sparse then
		numWonders = Round(numWonders * .75)
	elseif resource_abundance == abundance_values.Abundant then
		numWonders = Round(numWonders * 1.25)
	end
	
	--can only have one of each wonder and at least one wonder total
	--so basically, bound the number of wonders by 1 and all available wonders
	numWonders = math.max(math.min(numWonders, table.getn(natural_wonders)), 1)
	
	print ("Natural wonders to place: " .. numWonders)
	
	--organize all plots in the world by terrain types that wonders need
	local plot_terrains = {
		all = {}, 
		all_but_desert = {}, 
		grassland_plains = {}, 
		desert_plains = {}, 
		plains_desert_tundra = {},
		desert_tundra = {}
	}
	
	for y = 0, iH - 1 do
		for x = 0, iW - 1 do
			local plot = Map.GetPlot(x, y)
			local this_plot = {x = x, y = y}
			local this_plot_terrain = plot:GetTerrainType()
			
			table.insert(plot_terrains.all, this_plot)
			
			if this_plot_terrain == TerrainTypes.TERRAIN_GRASS then
				table.insert(plot_terrains.all_but_desert, this_plot)
				table.insert(plot_terrains.grassland_plains, this_plot)
			elseif this_plot_terrain == TerrainTypes.TERRAIN_PLAINS then
				table.insert(plot_terrains.all_but_desert, this_plot)
				table.insert(plot_terrains.grassland_plains, this_plot)
				table.insert(plot_terrains.desert_plains, this_plot)
				table.insert(plot_terrains.plains_desert_tundra, this_plot)
			elseif this_plot_terrain == TerrainTypes.TERRAIN_TUNDRA then
				table.insert(plot_terrains.all_but_desert, this_plot)
				table.insert(plot_terrains.plains_desert_tundra, this_plot)
				table.insert(plot_terrains.desert_tundra, this_plot)
			elseif this_plot_terrain == TerrainTypes.TERRAIN_SNOW then
				table.insert(plot_terrains.all_but_desert, this_plot)
			elseif this_plot_terrain == TerrainTypes.TERRAIN_DESERT then
				table.insert(plot_terrains.desert_plains, this_plot)
				table.insert(plot_terrains.plains_desert_tundra, this_plot)
				table.insert(plot_terrains.desert_tundra, this_plot)
			end
		end
	end
	
	--if the player has epic start on, then they get their own natural wonder
	local placed_wonders = {}
	if epic_start == 5 and not (Game.IsGameMultiPlayer() or Game.IsHotSeat()) then
		print("Placing Epic Start Wonder")
	
		--get all the plots within 1 to 4 distance of the player start
		local epicPlots = GetPlotsAtDistance(starting_points[1].x, starting_points[1].y, 1, 4)
		
		--pick a wonder
		local success
		local j = 0
		repeat
			local randWonder = natural_wonders[GetRandom(1, table.getn(natural_wonders))]
			success = false
			
			--print("attempting to place " .. randWonder.name)
			
			--randomize the order of plots
			epicPlots = GetShuffledCopyOfTable(epicPlots)
			for i = 1, table.getn(epicPlots) do
				epicPlot = Map.GetPlot(epicPlots[i].x, epicPlots[i].y)
				
				if randWonder.name == "potosi" then
					if PlaceWonderAtPlot(randWonder, epicPlots[i].x, epicPlots[i].y) == true then
						table.insert(placed_wonders, {name = randWonder.name, x = epicPlots[i].x, y = epicPlots[i].y})
						success = true
						break
					end
				elseif randWonder.name == "geyser" then
					if epicPlot:GetTerrainType() ~= TerrainTypes.TERRAIN_DESERT then
						if PlaceWonderAtPlot(randWonder, epicPlots[i].x, epicPlots[i].y) == true then
							table.insert(placed_wonders, {name = randWonder.name, x = epicPlots[i].x, y = epicPlots[i].y})
							success = true
							break
						end
					end
				elseif randWonder.name == "fountain" or randWonder.name == "eldorado" or randWonder.name == "sripada" or randWonder.name == "mtkailash" or randWonder.name == "mtfuji" then
					if epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_GRASS or epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_PLAINS then
						if PlaceWonderAtPlot(randWonder, epicPlots[i].x, epicPlots[i].y) == true then
							table.insert(placed_wonders, {name = randWonder.name, x = epicPlots[i].x, y = epicPlots[i].y})
							success = true
							break
						end
					end
				elseif randWonder.name == "mtsinai" or randWonder.name == "uluru" then
					if epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_DESERT or epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_PLAINS then
						if PlaceWonderAtPlot(randWonder, epicPlots[i].x, epicPlots[i].y) == true then
							table.insert(placed_wonders, {name = randWonder.name, x = epicPlots[i].x, y = epicPlots[i].y})
							success = true
							break
						end
					end
				elseif randWonder.name == "mesa" then
					if epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_DESERT or epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_PLAINS or epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_TUNDRA then
						if PlaceWonderAtPlot(randWonder, epicPlots[i].x, epicPlots[i].y) == true then
							table.insert(placed_wonders, {name = randWonder.name, x = epicPlots[i].x, y = epicPlots[i].y})
							success = true
							break
						end
					end
				elseif randWonder.name == "crater" then
					if epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_DESERT or epicPlot:GetTerrainType() == TerrainTypes.TERRAIN_TUNDRA then
						if PlaceWonderAtPlot(randWonder, epicPlots[i].x, epicPlots[i].y) == true then
							table.insert(placed_wonders, {name = randWonder.name, x = epicPlots[i].x, y = epicPlots[i].y})
							success = true
							break
						end
					end
				end
			end
			j = j + 1
		until success or j > 20
		
	end
	
	
	--now try placing wonders
	local wonder_dist = math.sqrt((iW * iH) / numWonders) / 2
	for i = 1, numWonders do
		local randWonder
		
		--can't be the same name. no doubles.
		local good_wonder
		local j = 0
		repeat
			randWonder = natural_wonders[GetRandom(1, table.getn(natural_wonders))]
			good_wonder = true
			for j = 1, table.getn(placed_wonders) do
				if placed_wonders[j].name == randWonder.name then
					good_wonder = false
					break
				end
			end
			j = j + 1
		until good_wonder == true or j > 50
		
		local wonder_plots = {}
		if randWonder.name == "potosi" then
			wonder_plots = plot_terrains.all
		elseif randWonder.name == "geyser" then
			wonder_plots = plot_terrains.all_but_desert
		elseif randWonder.name == "fountain" or randWonder.name == "eldorado" or randWonder.name == "sripada" or randWonder.name == "mtkailash" or randWonder.name == "mtfuji" then
			wonder_plots = plot_terrains.grassland_plains
		elseif randWonder.name == "mtsinai" or randWonder.name == "uluru" then
			wonder_plots = plot_terrains.desert_plains
		elseif randWonder.name == "mesa" then
			wonder_plots = plot_terrains.plains_desert_tundra
		elseif randWonder.name == "crater" then
			wonder_plots = plot_terrains.desert_tundra
		end
		
		for j = 1, 100 do --100 attempts
			local randPlot = GetRandom(1, table.getn(wonder_plots))
			randPlot = Map.GetPlot(wonder_plots[randPlot].x, wonder_plots[randPlot].y)
			
			--make sure the position jives with existing wonders
			local good_placement = true
			for k = 1, table.getn(placed_wonders) do
				if Map.PlotDistance(placed_wonders[k].x, placed_wonders[k].y, randPlot:GetX(), randPlot:GetY()) < wonder_dist then
					good_placement = false
					break
				end
			end
			
			if good_placement == true then
				if PlaceWonderAtPlot(randWonder, randPlot:GetX(), randPlot:GetY()) == true then
					table.insert(placed_wonders, {name = randWonder.name, x = randPlot:GetX(), y = randPlot:GetY()})
					break
				end
			end
		end
	end
end

function PlaceWonderAtPlot(wonder, x, y)
	local plot = Map.GetPlot(x, y)
	
	--make sure this isn't a player start. that could be bad.
	for i = 1, table.getn(starting_points) do
		if starting_points[i].x == x and starting_points[i].y == y then return false end
	end
	
	--make sure there are no resources here
	if plot:GetResourceType(-1) == -1 then
		--if this wonder requires a mountain, give it that
		if wonder.name == "geyser" or 
			wonder.name == "crater" or 
			wonder.name == "mtfuji" or 
			wonder.name == "mesa" or 
			wonder.name == "potosi" or 
			wonder.name == "sripada" or
			wonder.name == "mtsinai" or
			wonder.name == "mtkailash" or
			wonder.name == "uluru" 
		then
			plot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false)
		elseif wonder.name == "fountain" or wonder.name == "eldorado" then
			--else this wonder requires flat land
			plot:SetPlotType(PlotTypes.PLOT_LAND, false, false)
		end
		
		plot:SetFeatureType(wonder.id, -1)
		print ("Placed Natural Wonder " .. wonder.name .. " at " .. x .. ", " .. y)
		return true
	else
		return false
	end
end

function PlaceResources()
	abundance_values = {Sparse = 1, Normal = 2, Abundant = 3}
	resource_abundance = Map.GetCustomOption(5)
	
	local resources = {}
	
	for resource_data in GameInfo.Resources() do
		local resourceID = resource_data.ID;
		local resourceType = resource_data.Type;
		
		-- Set up Bonus IDs
		if resourceType == "RESOURCE_WHEAT" then
			resources["wheat"] = resourceID
		elseif resourceType == "RESOURCE_COW" then
			resources["cows"] = resourceID
		elseif resourceType == "RESOURCE_DEER" then
			resources["deer"] = resourceID
		elseif resourceType == "RESOURCE_BANANA" then
			resources["banana"] = resourceID
		elseif resourceType == "RESOURCE_FISH" then
			resources["fish"] = resourceID
		elseif resourceType == "RESOURCE_SHEEP" then
			resources["sheep"] = resourceID
		elseif resourceType == "RESOURCE_STONE" then
			resources["stone"] = resourceID
		-- Set up Strategic IDs
		elseif resourceType == "RESOURCE_IRON" then
			resources["iron"] = resourceID
		elseif resourceType == "RESOURCE_HORSE" then
			resources["horses"] = resourceID
		elseif resourceType == "RESOURCE_COAL" then
			resources["coal"] = resourceID
		elseif resourceType == "RESOURCE_OIL" then
			resources["oil"] = resourceID
		elseif resourceType == "RESOURCE_ALUMINUM" then
			resources["aluminum"] = resourceID
		elseif resourceType == "RESOURCE_URANIUM" then
			resources["uranium"] = resourceID
		-- Set up Luxury IDs
		elseif resourceType == "RESOURCE_WHALE" then
			resources["whales"] = resourceID
		elseif resourceType == "RESOURCE_PEARLS" then
			resources["pearls"] = resourceID
		elseif resourceType == "RESOURCE_IVORY" then
			resources["ivory"] = resourceID
		elseif resourceType == "RESOURCE_FUR" then
			resources["fur"] = resourceID
		elseif resourceType == "RESOURCE_SILK" then
			resources["silk"] = resourceID
		elseif resourceType == "RESOURCE_DYE" then
			resources["dye"] = resourceID
		elseif resourceType == "RESOURCE_SPICES" then
			resources["spices"] = resourceID
		elseif resourceType == "RESOURCE_SUGAR" then
			resources["sugar"] = resourceID
		elseif resourceType == "RESOURCE_COTTON" then
			resources["cotton"] = resourceID
		elseif resourceType == "RESOURCE_WINE" then
			resources["wine"] = resourceID
		elseif resourceType == "RESOURCE_INCENSE" then
			resources["incense"] = resourceID
		elseif resourceType == "RESOURCE_GOLD" then
			resources["gold"] = resourceID
		elseif resourceType == "RESOURCE_SILVER" then
			resources["silver"] = resourceID
		elseif resourceType == "RESOURCE_GEMS" then
			resources["gems"] = resourceID
		elseif resourceType == "RESOURCE_MARBLE" then
			resources["marble"] = resourceID
		-- Set up Expansion Pack Luxury IDs
		elseif resourceType == "RESOURCE_COPPER" then
			resources["copper"] = resourceID
		elseif resourceType == "RESOURCE_SALT" then
			resources["salt"] = resourceID
		elseif resourceType == "RESOURCE_CITRUS" then
			resources["citrus"] = resourceID
		elseif resourceType == "RESOURCE_TRUFFLES" then
			resources["truffles"] = resourceID
		elseif resourceType == "RESOURCE_CRAB" then
			resources["crabs"] = resourceID
		end
	end
	
	--organize our resources a little
	local strategic_resources = {resources.iron, resources.horses, resources.coal, resources.oil, resources.aluminum, resources.uranium}
	local food_resources = {resources.wheat, resources.cows, resources.deer, resources.banana, resources.sheep}
	local prod_resources = {resources.stone, resources.marble} --i like to put marble in here although it's technically a luxury
	local water_resources = {resources.fish, resources.pearls} --omit whales since there are only lakes in this map
	local prod_luxuries = {resources.gold, resources.silver, resources.gems} --these all give gold, but you have to mine them
	local gold_food_luxuries = {resources.ivory, resources.fur, resources.silk, resources.dye, resources.spices, resources.sugar, resources.cotton, resources.wine, resources.incense}
	
	--add G&K resources to their appropriate tables
	if isGAndK == true then
		table.insert(water_resources, resources.crabs)
		table.insert(prod_luxuries, resources.copper)
		table.insert(gold_food_luxuries, resources.salt)
		table.insert(gold_food_luxuries, resources.citrus)
		table.insert(gold_food_luxuries, resources.truffles)
	end
	
	--now we have to examine every plot in the map and organize its attributes to make it easier and quicker to add the resources
	local start_point_plots = {}
	local start_point_resources = {}
	
	--loop through each of the starting points' plots
	for i = 1, table.getn(starting_points) do
		start_point_plots[i] = {
			flat_plots = {}, 
			hill_plots = {}, 
			grassland_plots = {}, 
			plains_plots = {}, 
			tundra_plots = {}, 
			--snow_plots = {}, --unused
			desert_plots = {}, 
			forest_plots = {}, 
			jungle_plots = {}, 
			--marsh_plots = {}, --unused
			flat_grassland = {}, --no hills or features
			flat_plains = {}, --no hills or features
			flat_clear_plots = {}, --any terrain with no hills or features
			hill_clear_plots = {}, --hills with no features
			clear_plots = {}, --no features
			nonwasteland_plots = {} --anything but snow and desert
		}
		
		for j = 1, table.getn(starting_points[i].plots) do
			local x = starting_points[i].plots[j].x
			local y = starting_points[i].plots[j].y

			--organize the plot by terrain and features
			local plot = Map.GetPlot(x, y)
			local this_plot = starting_points[i].plots[j]
			local this_plot_type = plot:GetPlotType()
			local this_plot_terrain = plot:GetTerrainType()
			local this_plot_feature = plot:GetFeatureType()
			
			if this_plot_type ~= PlotTypes.PLOT_MOUNTAIN and  this_plot_type ~= PlotTypes.PLOT_OCEAN then
				if this_plot_type == PlotTypes.PLOT_LAND then
					table.insert(start_point_plots[i].flat_plots, this_plot)
					if this_plot_feature == FeatureTypes.NO_FEATURE then
						if this_plot_terrain == TerrainTypes.TERRAIN_GRASS then
							table.insert(start_point_plots[i].flat_grassland, this_plot)
						elseif this_plot_terrain == TerrainTypes.TERRAIN_PLAINS then
							table.insert(start_point_plots[i].flat_plains, this_plot)
						end
						table.insert(start_point_plots[i].flat_clear_plots, this_plot)
					end
				end
				if this_plot_type == PlotTypes.PLOT_HILLS then
					table.insert(start_point_plots[i].hill_plots, this_plot)
					if this_plot_feature == FeatureTypes.NO_FEATURE then
						table.insert(start_point_plots[i].hill_clear_plots, this_plot)
					end
				end
				--terrains
				if this_plot_terrain == TerrainTypes.TERRAIN_GRASS then
					table.insert(start_point_plots[i].grassland_plots, this_plot)
				elseif this_plot_terrain == TerrainTypes.TERRAIN_PLAINS then
					table.insert(start_point_plots[i].plains_plots, this_plot)
				elseif this_plot_terrain == TerrainTypes.TERRAIN_TUNDRA then
					table.insert(start_point_plots[i].tundra_plots, this_plot)
				--elseif this_plot_terrain == TerrainTypes.TERRAIN_SNOW then
					--table.insert(start_point_plots[i].snow_plots, this_plot)
				elseif this_plot_terrain == TerrainTypes.TERRAIN_DESERT then
					table.insert(start_point_plots[i].desert_plots, this_plot)
				end
				if this_plot_terrain ~= TerrainTypes.TERRAIN_DESERT and this_plot_terrain ~= TerrainTypes.TERRAIN_SNOW then
					table.insert(start_point_plots[i].nonwasteland_plots, this_plot)
				end
				--features
				if this_plot_feature == FeatureTypes.FEATURE_FOREST then
					table.insert(start_point_plots[i].forest_plots, this_plot)
				elseif this_plot_feature == FeatureTypes.FEATURE_JUNGLE then
					table.insert(start_point_plots[i].jungle_plots, this_plot)
				--elseif this_plot_feature == FeatureTypes.FEATURE_MARSH then
					--table.insert(start_point_plots[i].marsh_plots, this_plot)
				elseif this_plot_feature == FeatureTypes.NO_FEATURE then
					table.insert(start_point_plots[i].clear_plots, this_plot)
				end
			end
		end
		
		--deciding how many resources this start gets
		start_point_resources[i] = {}
		if starting_points[i].cityState == false then
			if resource_abundance == abundance_values.Sparse then
				start_point_resources[i].strategic_resource_num = GetRandom(2, 3)
				start_point_resources[i].food_resource_num = GetRandom(0, 1)
				start_point_resources[i].prod_resource_num = GetRandom(0, 1)
				start_point_resources[i].prod_luxuries_num = GetRandom(0, 1)
				start_point_resources[i].gold_food_luxuries_num = 1
			elseif resource_abundance == abundance_values.Normal then
				start_point_resources[i].strategic_resource_num = GetRandom(3, 4)
				start_point_resources[i].food_resource_num = GetRandom(1, 2)
				start_point_resources[i].prod_resource_num = GetRandom(1, 2)
				start_point_resources[i].prod_luxuries_num = GetRandom(0, 2)
				start_point_resources[i].gold_food_luxuries_num = GetRandom(1, 2)
			else --abundant
				start_point_resources[i].strategic_resource_num = GetRandom(4, 5)
				start_point_resources[i].food_resource_num = GetRandom(2, 3)
				start_point_resources[i].prod_resource_num = GetRandom(1, 3)
				start_point_resources[i].prod_luxuries_num = GetRandom(1, 2)
				start_point_resources[i].gold_food_luxuries_num = GetRandom(2, 3)
			end
		else --city-state
			if resource_abundance == abundance_values.Sparse then
				start_point_resources[i].strategic_resource_num = GetRandom(0, 1)
				start_point_resources[i].food_resource_num = GetRandom(0, 1)
				start_point_resources[i].prod_resource_num = GetRandom(0, 1)
				start_point_resources[i].prod_luxuries_num = GetRandom(0, 1)
				start_point_resources[i].gold_food_luxuries_num = GetRandom(0, 1)
			elseif resource_abundance == abundance_values.Normal then
				start_point_resources[i].strategic_resource_num = GetRandom(0, 2)
				start_point_resources[i].food_resource_num = GetRandom(0, 2)
				start_point_resources[i].prod_resource_num = GetRandom(0, 1)
				start_point_resources[i].prod_luxuries_num = GetRandom(0, 1)
				start_point_resources[i].gold_food_luxuries_num = GetRandom(0, 2)
			else --abundant
				start_point_resources[i].strategic_resource_num = GetRandom(1, 2)
				start_point_resources[i].food_resource_num = GetRandom(1, 2)
				start_point_resources[i].prod_resource_num = GetRandom(0, 1)
				start_point_resources[i].prod_luxuries_num = GetRandom(0, 1)
				start_point_resources[i].gold_food_luxuries_num = GetRandom(1, 2)
			end
		end
		
		--apply human start epicness resources
		if i == 1 and epic_start > 1 then
			--randomly give the player a couple more resources at low
			if epic_start >= 2 then
				for j = 1, 2 do
					local resource_type_num = {"strategic_resource_num", "food_resource_num", "prod_resource_num", "prod_luxuries_num", "gold_food_luxuries_num"}
					local randType = resource_type_num[GetRandom(1, 5)]
					start_point_resources[1][randType] = start_point_resources[1][randType] + 1
					print ("epic: extra  " .. randType)
				end
			end
			
			--give player more strategics at medium
			if epic_start >= 3 then
				start_point_resources[1].strategic_resource_num = start_point_resources[1].strategic_resource_num + 2
			end
			
			--give the player more resources at high
			if epic_start >= 4 then
				for j = 1, 2 do
					local resource_type_num = {"strategic_resource_num", "food_resource_num", "prod_resource_num", "prod_luxuries_num", "gold_food_luxuries_num"}
					local randType = resource_type_num[GetRandom(1, 5)]
					start_point_resources[1][randType] = start_point_resources[1][randType] + 1
					print ("epic: extra  " .. randType)
				end
			end
			
			--add to all resource types at epic
			if epic_start == 5 then
				--give player all strategics in good amounts
				for i = 1, table.getn(strategic_resources) do
					if strategic_resources[i] == resources.coal or strategic_resources[i] == resources.aluminum or strategic_resources[i] == resources.uranium then
						if GetRandom(1, 10) > 2 then
							PlaceResourceAtPlots(strategic_resources[i], start_point_plots[1].hill_plots, 1, 4, GetRandom(5, 9))
						else
							PlaceResourceAtPlots(strategic_resources[i], start_point_plots[1].flat_clear_plots, 1, 4, GetRandom(5, 9))
						end
					elseif strategic_resources[i] == resources.iron then
						if GetRandom(1, 10) > 2 then
							PlaceResourceAtPlots(resources.iron, start_point_plots[1].hill_plots, 1, 3, GetRandom(5, 9))
						else
							PlaceResourceAtPlots(resources.iron, start_point_plots[1].flat_clear_plots, 1, 3, GetRandom(5, 9))
						end
					elseif strategic_resources[i] == resources.horses then
						if GetRandom(1, 10) > 4 then
							PlaceResourceAtPlots(resources.horses, start_point_plots[1].flat_grassland, 1, 3, GetRandom(5, 9))
						else
							PlaceResourceAtPlots(resources.horses, start_point_plots[1].flat_plains, 1, 3, GetRandom(5, 9))
						end
					elseif strategic_resources[i] == resources.oil then
						if GetRandom(1, 10) > 3 then
							PlaceResourceAtPlots(resources.oil, start_point_plots[1].flat_plots, 1, 4, GetRandom(5, 9))
						else
							PlaceResourceAtPlots(resources.oil, start_point_plots[1].hill_plots, 1, 4, GetRandom(5, 9))
						end
					end
				end
				start_point_resources[1].strategic_resource_num = GetRandom(2, 4) --a few more strategics for the road
				start_point_resources[1].food_resource_num = start_point_resources[1].food_resource_num + 1
				start_point_resources[1].prod_resource_num = start_point_resources[1].prod_resource_num + 1
				start_point_resources[1].prod_luxuries_num = start_point_resources[1].prod_luxuries_num + 1
				start_point_resources[1].gold_food_luxuries_num = start_point_resources[1].gold_food_luxuries_num + 1
			end
		end
		
		--init resource counters
		start_point_resources[i].counts = {}
		for key, value in pairs(resources) do 
			start_point_resources[i].counts[value] = 0
		end
	end
	
	--now we have enough information to finally start placing these bitches
	for i = 1, table.getn(starting_points) do
		--[[
		print("Player " .. i .. " strategic resources: " .. start_point_resources[i].strategic_resource_num)
		print("Player " .. i .. " food resources: " .. start_point_resources[i].food_resource_num)
		print("Player " .. i .. " production resources: " .. start_point_resources[i].prod_resource_num)
		print("Player " .. i .. " production luxuries: " .. start_point_resources[i].prod_luxuries_num)
		print("Player " .. i .. " luxury resources: " .. start_point_resources[i].gold_food_luxuries_num)
		]]
	
		--place strategics
		local strategic_resources_placed = 0
		local max_attempts = table.getn(strategic_resources) * 2
		local strategic_attempts = 0
		while strategic_resources_placed < start_point_resources[i].strategic_resource_num and strategic_attempts < max_attempts do 
			--pick a random strategic
			local randRes
			local success = false
			local loops = 1 --always put in a max loop counter or you're gonna have a bad time
			repeat
				randRes = strategic_resources[GetRandom(1, table.getn(strategic_resources))]
				loops = loops + 1
			until start_point_resources[i].counts[randRes] < 2 or loops > 25 --don't place more than two of the same resource
			
			strategic_attempts = strategic_attempts + 1
			
			if randRes == resources.coal or randRes == resources.aluminum or randRes == resources.uranium then
				--80% chance of hill placement, 20% any flat clear
				if GetRandom(1, 10) > 2 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].hill_plots, 1, 5, GetRandom(3, 9))
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_clear_plots, 1, 5, GetRandom(3, 9))
				end
			elseif randRes == resources.iron then
				--80% chance of hill placement, 20% any flat clear
				if GetRandom(1, 10) > 2 then
					success = PlaceResourceAtPlots(resources.iron, start_point_plots[i].hill_plots, 1, 4, GetRandom(3, 9))
				else
					success = PlaceResourceAtPlots(resources.iron, start_point_plots[i].flat_clear_plots, 1, 4, GetRandom(3, 9))
				end
			elseif randRes == resources.horses then
				--either grassland or plains will do, but slight preference to grassland
				if GetRandom(1, 10) > 4 then
					success = PlaceResourceAtPlots(resources.horses, start_point_plots[i].flat_grassland, 1, 4, GetRandom(3, 9))
				else
					success = PlaceResourceAtPlots(resources.horses, start_point_plots[i].flat_plains, 1, 4, GetRandom(3, 9))
				end
			elseif randRes == resources.oil then
				--flat land mostly, but could be in hills
				if GetRandom(1, 10) > 3 then
					success = PlaceResourceAtPlots(resources.oil, start_point_plots[i].flat_plots, 1, 5, GetRandom(3, 9))
				else
					success = PlaceResourceAtPlots(resources.oil, start_point_plots[i].hill_plots, 1, 5, GetRandom(3, 9))
				end
			end
			
			--increment the counter on this resource
			if success == true then 
				start_point_resources[i].counts[randRes] = start_point_resources[i].counts[randRes] + 1
				strategic_resources_placed = strategic_resources_placed + 1
				--print("placed strategic resource " .. randRes)
			end
		end
		
		--place food resources
		local food_resources_placed = 0
		max_attempts = table.getn(food_resources) * 2
		local food_attempts = 0
		while food_resources_placed < start_point_resources[i].food_resource_num and food_attempts < max_attempts do 
			--pick a random resource
			local randRes
			local success = false
			local loops = 1
			repeat
				randRes = food_resources[GetRandom(1, table.getn(food_resources))]
				loops = loops + 1
			until start_point_resources[i].counts[randRes] < 2 or loops > 25 --don't place more than two of the same resource
			
			food_attempts = food_attempts + 1
			
			if randRes == resources.wheat then
				--typically, wheat only goes on flat plains, but i'm going to allow some on flat grassland, too
				if GetRandom(1, 10) > 3 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_plains, 1, 2)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_grassland, 1, 2)
				end
			elseif randRes == resources.cows then
				--cows are mostly on grassland, but i'm going to allow some on plains
				if GetRandom(1, 10) > 3 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_grassland, 1, 2)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_plains, 1, 2)
				end
			elseif randRes == resources.deer then
				--deer only do forest
				success = PlaceResourceAtPlots(randRes, start_point_plots[i].forest_plots, 1, 3)
			elseif randRes == resources.banana then
				--bananas only do jungle
				success = PlaceResourceAtPlots(randRes, start_point_plots[i].jungle_plots, 1, 2)
			elseif randRes == resources.sheep then
				--sheep prefer hills but will go anywhere clear, really. those sheep whores.
				if GetRandom(1, 10) > 3 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].hill_clear_plots, 1, 3)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_plains, 2, 4)
				end
			end
			
			--increment the counter on this resource
			if success == true then 
				start_point_resources[i].counts[randRes] = start_point_resources[i].counts[randRes] + 1
				food_resources_placed = food_resources_placed + 1
				--print("placed food resource " .. randRes)
			else
				--print ("placing food resource " .. randRes .. " failed")
			end
		end
		
		--place production resources
		local prod_resources_placed = 0
		max_attempts = table.getn(prod_resources) * 2
		local prod_attempts = 0
		while prod_resources_placed < start_point_resources[i].prod_resource_num and prod_attempts < max_attempts do 
			--pick a random resource
			local randRes
			local success = false
			local loops = 1
			repeat
				randRes = prod_resources[GetRandom(1, table.getn(prod_resources))]
				loops = loops + 1
			until start_point_resources[i].counts[randRes] < 2 or loops > 12 --don't place more than two of the same resource
			
			prod_attempts = prod_attempts + 1
			
			if randRes == resources.stone then
				--70% chance of hill placement, 30% any flat clear
				if GetRandom(1, 10) > 3 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].hill_plots, 1, 3)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_clear_plots, 1, 3)
				end
			elseif randRes == resources.marble then
				--70% chance of hill placement, 30% any flat clear
				if GetRandom(1, 10) > 3 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].hill_plots, 1, 4)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_clear_plots, 1, 4)
				end
			end
			
			--increment the counter on this resource
			if success == true then 
				start_point_resources[i].counts[randRes] = start_point_resources[i].counts[randRes] + 1
				prod_resources_placed = prod_resources_placed + 1
				--print("placed production resource " .. randRes)
			end
		end
		
		--place production luxuries
		local prod_luxuries_placed = 0
		max_attempts = table.getn(prod_luxuries) * 2
		local prod_luxuries_attempts = 0
		while prod_luxuries_placed < start_point_resources[i].prod_luxuries_num and prod_luxuries_attempts < max_attempts do 
			--pick a random luxury
			local randRes
			local success = false
			local loops = 1
			repeat
				randRes = prod_luxuries[GetRandom(1, table.getn(prod_luxuries))]
				loops = loops + 1
			until start_point_resources[i].counts[randRes] < 2 or loops > 25 --don't place more than two of the same resource
			
			prod_luxuries_attempts = prod_luxuries_attempts + 1
			
			if randRes == resources.gold or randRes == resources.silver or randRes == resources.gems or randRes == resources.copper then
				--80% chance of hill placement, 20% anywhere
				if GetRandom(1, 10) > 2 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].hill_plots, 1, 4)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].nonwasteland_plots, 1, 4)
				end
			end
			
			--increment the counter on this resource
			if success == true then 
				start_point_resources[i].counts[randRes] = start_point_resources[i].counts[randRes] + 1
				prod_luxuries_placed = prod_luxuries_placed + 1
				--print("placed production luxury " .. randRes)
			end
		end
		
		--place gold luxuries
		local gold_food_luxuries_placed = 0
		max_attempts = table.getn(gold_food_luxuries) * 2
		local gold_food_luxuries_attempts = 0
		while gold_food_luxuries_placed < start_point_resources[i].gold_food_luxuries_num and gold_food_luxuries_attempts < max_attempts do 
			--pick a random luxury
			local randRes
			local success = false
			local loops = 1
			repeat
				randRes = gold_food_luxuries[GetRandom(1, table.getn(gold_food_luxuries))]
				loops = loops + 1
			until start_point_resources[i].counts[randRes] < 2 or loops > 25 --don't place more than two of the same resource
			
			gold_food_luxuries_attempts = gold_food_luxuries_attempts + 1
			
			if randRes == resources.salt then
				--salt goes anywhere clear
				success = PlaceResourceAtPlots(randRes, start_point_plots[i].clear_plots, 1, 5)
			elseif randRes == resources.citrus then
				--citrus only does jungle
				success = PlaceResourceAtPlots(randRes, start_point_plots[i].jungle_plots, 1, 5)
			elseif randRes == resources.truffles then
				--truffles can be anywhere except wasteland
				success = PlaceResourceAtPlots(randRes, start_point_plots[i].nonwasteland_plots, 1, 5)
			elseif randRes == resources.ivory or randRes == resources.sugar then
				--elephants like clear, flat grassland
				success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_grassland, 1, 5)
			elseif randRes == resources.fur then
				--furs can be found mainly in forests but also tundra or plains
				if GetRandom(1, 10) > 4 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].forest_plots, 1, 5)
				elseif GetRandom(1, 10) > 5 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].plains_plots, 1, 5)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].tundra_plots, 1, 5)
				end
			elseif randRes == resources.silk or randRes == resources.dye or randRes == resources.spices or randRes == resources.wine then
				--many luxuries like grassland or plains
				if GetRandom(1, 10) > 4 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].grassland_plots, 1, 5)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].plains_plots, 1, 5)
				end
			elseif randRes == resources.cotton then
				--cotton prefers open grassland or plains
				if GetRandom(1, 10) > 5 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_grassland, 1, 5)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].flat_plains, 1, 5)
				end
			elseif randRes == resources.incense then
				--incense can be plains or desert
				if GetRandom(1, 10) > 5 then
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].plains_plots, 1, 5)
				else
					success = PlaceResourceAtPlots(randRes, start_point_plots[i].desert_plots, 1, 5)
				end
			end
			
			--increment the counter on this resource
			if success == true then 
				start_point_resources[i].counts[randRes] = start_point_resources[i].counts[randRes] + 1
				gold_food_luxuries_placed = gold_food_luxuries_placed + 1
				--print("placed luxury resource " .. randRes)
			end
		end
	end
	
	--lastly, we place the water resources because the lakes were already generated
	--sucks that you can't improve these resources but at least they make the lakes better
	for i = 1, table.getn(lakes) do
		local numLakeResources = 0
		local lakeSize = table.getn(lakes[i])
		
		if lakeSize == 1 then
			--70% chance of a resource
			if GetRandom(1, 10) <= 7 then
				numLakeResources = 1
			end
		elseif lakeSize == 2 then
			--90% chance of one resource, 40% of two
			local randResources = GetRandom(1, 10)
			if randResources <= 4 then
				numLakeResources = 2
			elseif randResources <= 9 then
				numLakeResources = 1
			end
		elseif lakeSize == 3 then
			--100% chance of one resource, 60% of two, 10% of three
			local randResources = GetRandom(1, 10)
			if randResources == 1 then
				numLakeResources = 3
			elseif randResources <= 6 then
				numLakeResources = 2
			else
				numLakeResources = 1
			end
		else --lakesize >= 4
			--100% chance of one resource, 70% of two, 30% of three
			local randResources = GetRandom(1, 10)
			if randResources <= 3 then
				numLakeResources = 3
			elseif randResources <= 7 then
				numLakeResources = 2
			else
				numLakeResources = 1
			end
		end
		
		
		repeat
			local randRes = water_resources[GetRandom(1, table.getn(water_resources))]
			local randLakePlot = lakes[i][GetRandom(1, lakeSize)]
			randLakePlot = Map.GetPlot(randLakePlot.x, randLakePlot.y)
			
			if randLakePlot:GetResourceType(-1) == -1 then
				randLakePlot:SetResourceType(randRes, 1)
				numLakeResources = numLakeResources - 1
			end
		until numLakeResources < 1 --there's no way to get an infinite loop here
	end
	
	--[[
	for i = 1, table.getn(start_point_plots) do
		for key,value in pairs(start_point_plots[i]) do 
			print(i .. "." .. key .. ": " .. table.getn(value))
		end
	end
	]]
end

--attempts to place a resource on a random plot from a list of plots
--resource id, table of plots, amount of resources (strategics)
function PlaceResourceAtPlots(resource, plots, min_depth, max_depth, amount)
	if amount == nil then amount = 1 end
	if min_depth == nil then min_depth = 0 end
	if max_depth == nil then max_depth = 99 end
	if table.getn(plots) == 0 then return false end

	local k = 1
	local success = false
	repeat
		local randomPlot = GetRandom(1, table.getn(plots))
		local plot = Map.GetPlot(plots[randomPlot].x, plots[randomPlot].y)
		
		--make sure this isn't a player start. setting a city up on a resource isn't preferable
		local on_player_start = false
		for i = 1, table.getn(starting_points) do
			if starting_points[i].x == plots[randomPlot].x and starting_points[i].y == plots[randomPlot].y then 
				on_player_start = true
				break
			end
		end
		
		--no existing resource
		--tile is passable
		--tile is between min and max depth
		if plot:GetResourceType(-1) == -1 and plot:IsImpassable() == false and on_player_start == false and plots[randomPlot].depth >= min_depth and plots[randomPlot].depth <= max_depth then
			--print ("  placed " .. resource .. " at depth " .. plots[randomPlot].depth)
			plot:SetResourceType(resource, amount)
			success = true
		end
		
		k = k + 1
	until k > table.getn(plots) or success == true --loop until we succeed or exhaust the number of plots
	
	return success
end

function AddGoodies()
	local goodies = {}
	
	--give each starting point a few goodies
	for i = 1, table.getn(starting_points) do
		if starting_points[i].cityState == false then
			local numGoodies = GetRandom(2, 3)
			
			--apply epic bonus goody
			if i == 1 and epic_start > 1 then numGoodies = numGoodies + 1 end
			
			for j = 1, numGoodies do
				for k = 1, 20 do
					local goodyDiameter = (player_size * 2) + 5
					local goodyRadius = math.floor(goodyDiameter / 2)
					local randX = (GetRandom(0, goodyDiameter) - goodyRadius) + starting_points[i].x
					local randY = (GetRandom(0, goodyDiameter) - goodyRadius) + starting_points[i].y
					local plot = Map.GetPlot(randX, randY)
					
					if plot ~= nil then
						if plot:GetPlotType() ~= PlotTypes.PLOT_MOUNTAIN and plot:GetPlotType() ~= PlotTypes.PLOT_OCEAN and plot:GetResourceType(-1) == -1 then
							local good_placement = true
							--enforce min and max distance from starting point
							local goodyDepth = Map.PlotDistance(starting_points[i].x, starting_points[i].y, randX, randY)
							if goodyDepth < 2 or goodyDepth > goodyRadius then
								good_placement = false
							end
							
							--distance from other goodies
							if good_placement == true then
								for l=1, table.getn(goodies) do
									if (Map.PlotDistance(goodies[l].x, goodies[l].y, randX, randY) < player_size) then
										good_placement = false
										break
									end
								end
							end
							
							if good_placement == true then
								--plot:SetFeatureType(FeatureTypes.FEATURE_FALLOUT, -1)
								plot:SetImprovementType(GameInfo.Improvements.IMPROVEMENT_GOODY_HUT.ID);
								table.insert(goodies, {x = randX, y = randY})
								break
							end
						end
					end
				end
			end
		end
	end
	print ("Player Goodies Placed: " .. table.getn(goodies))
	
	--now place some more goodies out in the wastes
	local free_space = (iW * iH) - (iNumCivs * GetNumberOfTiles(player_size)) - (iNumCityStates * GetNumberOfTiles(cs_size))
	local numGoodies = math.floor(iNumCivs * 1.5) -- each player gets 1 or 2 more goodies--whoever gets there first ;)
	local goody_dist = math.max(math.ceil(math.sqrt(free_space / numGoodies) / 2), 5) --minimum distance between goodies
	
	print ("Total Attempted Wasteland Goodies: " .. numGoodies)
	
	for i = 1, numGoodies do
		for j = 1, 20 do
			local randX = GetRandom(0, iW - 1)
			local randY = GetRandom(0, iH - 1)
			local plot = Map.GetPlot(randX, randY)
			
			if plot:GetPlotType() ~= PlotTypes.PLOT_MOUNTAIN and plot:GetPlotType() ~= PlotTypes.PLOT_OCEAN and (plot:GetTerrainType() == TerrainTypes.TERRAIN_SNOW or plot:GetTerrainType() == TerrainTypes.TERRAIN_DESERT) then
				local good_placement = true
				for k=1, table.getn(goodies) do
					if (Map.PlotDistance(goodies[k].x, goodies[k].y, randX, randY) < goody_dist) then
						good_placement = false
						break
					end
				end
				
				if good_placement == true then
					--plot:SetFeatureType(FeatureTypes.FEATURE_ICE, -1)
					plot:SetImprovementType(GameInfo.Improvements.IMPROVEMENT_GOODY_HUT.ID);
					table.insert(goodies, {x = randX, y = randY})
					break
				end
			end
		end
	end
	print ("Total Goodies Placed: " .. table.getn(goodies))
end

function SetContinentArt()
	for i, plot in Plots() do
		if plot:IsWater() then
			plot:SetContinentArtType(0);
		else
			if Map.GetCustomOption(1) == 1 then
				plot:SetContinentArtType(4);
			else
				plot:SetContinentArtType(3);
			end
		end
	end
end

function PlaceOddities()
	local oddityLevel = Map.GetCustomOption(7)
	if oddityLevel > 1 then
		--figure out how many oddities to place
		local sizevalues = {
			[0] = .5, --WORLDSIZE_DUEL
			[1] = .75, --WORLDSIZE_TINY
			[2] = 1, --WORLDSIZE_SMALL
			[3] = 1.5, --WORLDSIZE_STANDARD
			[4] = 2, --WORLDSIZE_LARGE
			[5] = 2.5 --WORLDSIZE_HUGE
		};
		local size_multiplier = sizevalues[Map.GetWorldSize()] or 1;
		local oddity_number = {
			[1] = 0,
			[2] = 2,
			[3] = 3,
			[4] = 4,
			[5] = 4,
			[6] = 5
		}
		
		local numOddities = math.max(math.ceil(oddity_number[oddityLevel] * size_multiplier), 1)
		
		local oddities = {
			{name = "fort", clearSpace = 0, level = 2}, --a hill with a fort on it
			{name = "ruins", clearSpace = 3, level = 2}, --a city ruin with roads passing through it
			{name = "lush", clearSpace = 3, level = 3}, --a small patch of fertile land (~7 tiles)
			{name = "oasis", clearSpace = 2, level = 3}, --a barbarian worker at an oasis (with desert under it)
			{name = "great", clearSpace = 1, level = 4}, --a random great person site
			{name = "goody", clearSpace = 1, level = 4}, --a goody hut surrounded by mountains
			{name = "moai", clearSpace = 2, level = 5}, --an ocean plot surrounded by moai
			{name = "fallout", clearSpace = 3, level = 5}, --rich uranium deposit surrounded by fallout
			{name = "uberbarb", clearSpace = 0, level = 6}, --a barbarian unit that is a couple eras ahead of starting era
			{name = "dong", clearSpace = 4, level = 6} --hill formation shaped like a dong, with old faithful at the ti
		}
		
		--figure out what oddities are available at this level
		local availableOddities = {}
		for i = 1, table.getn(oddities) do 
			if oddities[i].level <= oddityLevel then
				if oddities[i].name == "uberbarb" then
					if Game.GetCustomOption("GAMEOPTION_NO_BARBARIANS") == 0 then
						table.insert(availableOddities, oddities[i])
					end
				elseif oddities[i].name == "moai" then
					if GameInfo.Improvements.IMPROVEMENT_MOAI ~= nil then
						table.insert(availableOddities, oddities[i])
					end
				else
					table.insert(availableOddities, oddities[i])
				end
			end
		end
		
		local defaultTerrain
		if Map.GetCustomOption(1) == 1 then
			defaultTerrain = TerrainTypes.TERRAIN_SNOW
		else
			defaultTerrain = TerrainTypes.TERRAIN_DESERT
		end
		
		local existing_oddities = {}
		
		for i = 1, numOddities do
		
			--decide what oddity to try to place
			local oddity = availableOddities[GetRandom(1, table.getn(availableOddities))]
			print ("attempting to place oddity: " .. oddity.name)
			
			--now randomly find a place to put it
			--i figure this must be faster than indexing every tile
			local max_attempts = math.floor((iW * iH) / 10)
			for j = 1, max_attempts do
				local randX = GetRandom(0, iW - 1)
				local randY = GetRandom(0, iH - 1)
				local plot = Map.GetPlot(randX, randY)
				
				--make sure it's at least wasteland
				if plot:GetTerrainType() == defaultTerrain then
					--all oddities must be at least 2 + size tiles from a civ, 1 + size from a city state
					local good_placement = true
					for k = 1, table.getn(starting_points) do
						local oddity_dist;
						if starting_points[k].cityState == false then
							oddity_dist = player_size + oddity.clearSpace + 2
						else
							oddity_dist = cs_size + oddity.clearSpace + 1
						end
						
						if Map.PlotDistance(starting_points[k].x, starting_points[k].y, randX, randY) < oddity_dist then
							good_placement = false
							break
						end
					end
					
					
					if good_placement == true then
						--make sure we're not within any other oddity's clear space
						if table.getn(existing_oddities) > 0 then
							for k = 1, table.getn(existing_oddities) do
								if Map.PlotDistance(existing_oddities[k].x, existing_oddities[k].y, randX, randY) < existing_oddities[k].clearSpace + oddity.clearSpace then
									good_placement = false
									break
								end
							end
						end
						
						if good_placement == true then
							--now that we know the plot is away from all starts and existing oddities, 
							--we can check that the required clear land is available
							if oddity.clearSpace > 0 then
								local plots = GetPlotsAtDistance(randX, randY, 1, oddity.clearSpace)
								
								if plots ~= nil then
									for k = 1, table.getn(plots) do
										local adjPlot = Map.GetPlot(plots[k].x, plots[k].y)
										if adjPlot:GetTerrainType() ~= defaultTerrain or adjPlot:GetPlotType() ~= PlotTypes.PLOT_LAND then
											good_placement = false
											break
										end
									end
								else
									good_placement = false
								end
							end
							
							if good_placement == true then
								--hurray! now we can place the oddity here!
								if oddity.name == "fort" then
									plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
									plot:SetImprovementType(GameInfo.Improvements.IMPROVEMENT_FORT.ID);
								elseif oddity.name == "ruins" then
									plot:SetImprovementType(GameInfo.Improvements.IMPROVEMENT_CITY_RUINS.ID);
									plot:SetRouteType(GameInfo.Routes.ROUTE_ROAD.ID)
									--add 2 to 4 roads coming from the ruins
									local numRoads = GetRandom(2, 4)
									local roadDir = GetRandom(0, 5)
									
									for k = 1, numRoads do
										local roadLength = GetRandom(2, 3)
										local nextPlot = plot
										for l = 1, roadLength do
											nextPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), roadDir)
											if nextPlot == nil then break end
											nextPlot:SetRouteType(GameInfo.Routes.ROUTE_ROAD.ID)
										end
										roadDir = math.ceil((roadDir + (6 / numRoads)) % 6)
									end
								elseif oddity.name == "lush" then
									local plots = GetPlotsAtDistance(randX, randY, 0, 2)
									local junglePlots = {}
									if plots ~= nil then
										for k = 1, table.getn(plots) do
											local thisPlot = Map.GetPlot(plots[k].x, plots[k].y)
											
											local lushIt = false
											if plots[k].depth == 0 then
												 lushIt = true
											elseif plots[k].depth == 1 then
												if GetRandom(1, 6) <= 5 then
													lushIt = true
												end
											else
												if GetRandom(1, 6) <= 2 then
													lushIt = true
												end
											end
											
											if lushIt == true then
												--all grassland
												thisPlot:SetTerrainType(TerrainTypes.TERRAIN_GRASS, false, false)
												--give it some hills
												if GetRandom(1, 10) <= 4 then
													thisPlot:SetPlotType(PlotTypes.PLOT_HILLS, false, false)
												end
												--mostly jungle
												if GetRandom(1, 10) <= 6 then
													thisPlot:SetFeatureType(FeatureTypes.FEATURE_JUNGLE, -1)
													table.insert(junglePlots, {
														x = thisPlot:GetX(), 
														y = thisPlot:GetY(), 
														depth = Map.PlotDistance(randX, randY, thisPlot:GetX(), thisPlot:GetY())
													})
												end
											end
										end
										--give a couple bananas
										PlaceResourceAtPlots(GameInfo.Resources.RESOURCE_BANANA.ID, junglePlots, 0, 1)
										PlaceResourceAtPlots(GameInfo.Resources.RESOURCE_BANANA.ID, junglePlots, 0, 1)
									end
								elseif oddity.name == "moai" then
									plot:SetPlotType(PlotTypes.PLOT_OCEAN, false, false)
									if GetRandom(1, 10) > 6 and is_worldbuilder == false then
										local battleship = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_BATTLESHIP"], randX, randY)
										battleship:SetName("This is going to kill you!")
									end
									local plots = GetPlotsAtDistance(randX, randY, 0, 1)
									if plots ~= nil then
										for k = 1, table.getn(plots) do
											local thisPlot = Map.GetPlot(plots[k].x, plots[k].y)
											thisPlot:SetImprovementType(GameInfo.Improvements.IMPROVEMENT_MOAI.ID)
										end
									end
								elseif oddity.name == "great" then
									local greatPeopleThings = {
										GameInfo.Improvements.IMPROVEMENT_LANDMARK.ID, 
										GameInfo.Improvements.IMPROVEMENT_ACADEMY.ID, 
										GameInfo.Improvements.IMPROVEMENT_CUSTOMS_HOUSE.ID, 
										GameInfo.Improvements.IMPROVEMENT_MANUFACTORY.ID,
									}
									if isGAndK == true then
										greatPeopleThings[5] = GameInfo.Improvements.IMPROVEMENT_HOLY_SITE.ID
									end
									
									if defaultTerrain == TerrainTypes.TERRAIN_SNOW then
										plot:SetTerrainType(TerrainTypes.TERRAIN_DESERT, false, false)
									else
										plot:SetTerrainType(TerrainTypes.TERRAIN_SNOW, false, false)
									end
									plot:SetImprovementType(greatPeopleThings[GetRandom(1, 5)])
								elseif oddity.name == "oasis" then
									plot:SetTerrainType(TerrainTypes.TERRAIN_DESERT, false, false)
									plot:SetFeatureType(FeatureTypes.FEATURE_OASIS, -1) --crashes if not on desert
									--next line crashes civ when running through worldbuilder.
									if is_worldbuilder == false then
										local worker = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_WORKER"], plot:GetX(), plot:GetY(), 0) --no ai
										local workerNames = {"Dave", "Jerry", "Mr. Larrity", "Benny", "Todd", "Black Steve", "Clare", "Mary", "Dean", "Clarence"}
										worker:SetName(workerNames[GetRandom(1, table.getn(workerNames))])
									end
								elseif oddity.name == "fallout" then
									plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
									plot:SetResourceType(GameInfo.Resources.RESOURCE_URANIUM.ID, GetRandom(12, 99))
									plot:SetFeatureType(FeatureTypes.FEATURE_FALLOUT, -1)
									local plots = GetPlotsAtDistance(randX, randY, 1, 2)
									if plots ~= nil then
										for k = 1, table.getn(plots) do
											local thisPlot = Map.GetPlot(plots[k].x, plots[k].y)
											if plots[k].depth == 1 then
												--chance to be a hill
												if GetRandom(1, 6) <= 4 then
													thisPlot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
												end
												--chance to have fallout
												if GetRandom(1, 6) <= 5 then
													thisPlot:SetFeatureType(FeatureTypes.FEATURE_FALLOUT, -1)
												end
											else
												--chance to be a hill
												if GetRandom(1, 12) <= 2 then
													thisPlot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
												end
												--chance to have fallout
												if GetRandom(1, 12) <= 2 then
													thisPlot:SetFeatureType(FeatureTypes.FEATURE_FALLOUT, -1)
												end
											end
										end
									end
								elseif oddity.name == "goody" then
									plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false)
									plot:SetTerrainType(TerrainTypes.TERRAIN_GRASS, false, false)
									plot:SetImprovementType(GameInfo.Improvements.IMPROVEMENT_GOODY_HUT.ID)
									local randDir = GetRandom(0, 5)
									for i = 0, 5 do
										local nextPlot = Map.PlotDirection(plot:GetX(), plot:GetY(), i)
										if nextPlot == nil then break end
										if i == randDir then
											nextPlot:SetPlotType(PlotTypes.PLOT_HILLS, false, false)
										else
											nextPlot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false)
										end
									end
								elseif oddity.name == "uberbarb" then
									if is_worldbuilder == false then
										local uberBarb
										if Game.GetStartEra() == GameInfo.Eras["ERA_ANCIENT"].ID then
											uberBarb = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_JAPANESE_SAMURAI"], randX, randY)
										elseif Game.GetStartEra() == GameInfo.Eras["ERA_CLASSICAL"].ID then
											uberBarb = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_FRENCH_MUSKETEER"], randX, randY)
										elseif Game.GetStartEra() == GameInfo.Eras["ERA_MEDIEVAL"].ID then
											uberBarb = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_RUSSIAN_COSSACK"], randX, randY)
										elseif Game.GetStartEra() == GameInfo.Eras["ERA_RENAISSANCE"].ID then
											uberBarb = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_RUSSIAN_COSSACK"], randX, randY)
										elseif Game.GetStartEra() == GameInfo.Eras["ERA_INDUSTRIAL"].ID then
											uberBarb = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_GERMAN_PANZER"], randX, randY)
										elseif Game.GetStartEra() == GameInfo.Eras["ERA_MODERN"].ID or Game.GetStartEra() == GameInfo.Eras["ERA_FUTURE"].ID then 
											uberBarb = Players[GameDefines.MAX_CIV_PLAYERS]:InitUnit(GameInfoTypes["UNIT_MECH"], randX, randY)
										end
										
										if uberBarb ~= nil then
											local uberBarbNames = {"Mr. Papadopoulos", "Chewbacca", "Tacocat", "Scruffy", "Ace Rimmer", "Vic Romano", "Sheldor the Conqueror", "Beavis", "Master Shake", "Conan"}
											uberBarb:SetName(uberBarbNames[GetRandom(1, table.getn(uberBarbNames))])
										end
									end
								elseif oddity.name == "dong" then
									local nextPlot = plot
									dongDir = GetRandom(1, 2)
									if dongDir == 1 then
										dongDir = DirectionTypes.DIRECTION_NORTHEAST
									else
										dongDir = DirectionTypes.DIRECTION_NORTHWEST
									end
									
									--3 hill plots for shaft
									for k = 1, 3 do
										nextPlot:SetPlotType(PlotTypes.PLOT_HILLS, false, false)
										if k == 3 then
											--old faithful at the tip!!
											nextPlot:SetFeatureType(GameInfo.Features.FEATURE_GEYSER.ID, -1)
										else
											nextPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), dongDir)
											if nextPlot == nil then break end
										end
									end
									
									--2 more plots for the shaft in the opposite direction
									nextPlot = plot
									dongDir = (dongDir + 3) % 6
									for k = 1, 2 do
										nextPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), dongDir)
										if nextPlot == nil then break end
										nextPlot:SetPlotType(PlotTypes.PLOT_HILLS, false, false)
									end
									
									--two mountains for balls at the base!
									local ballPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), (dongDir + 1) % 6)
									if ballPlot ~= nil then
										ballPlot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false)
									end
									ballPlot = Map.PlotDirection(nextPlot:GetX(), nextPlot:GetY(), (dongDir - 1) % 6)
									if ballPlot ~= nil then
										ballPlot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false)
									end
								end
								
								table.insert(existing_oddities, {x = randX, y = randY, clearSpace = oddity.clearSpace})
								print("   placed oddity | attempts: " .. j)
								--successful placement so break this loop
								break
							end
						end
					end
				end
				if j == max_attempts then print("   failed to placed | attempts: " .. max_attempts) end
			end --end j
		end --end i
	end
end